import pytest

from ..pai_client import PAIClient


def test_initialization_with_auth():
    client = PAIClient("http", "localhost", "8080", api_key="test")
    assert client.get.headers["x-api-key"] == "test"
    client = PAIClient("http", "localhost", "8080", bearer_token="test")
    assert client.get.headers["Authorization"] == "Bearer test"


def test_initialization_error_message():
    with pytest.raises(ValueError) as e:
        client = PAIClient()

    assert e.match(
        "PAIClient needs either a url, or a scheme and host to initialize. You can find more information on which url to use here: https://docs.private-ai.com/thin-client/"
    )



from requests import HTTPError, Response

from .request_objects import Entity, ReidentifyTextRequest


class BaseResponse:
    def __init__(self, response_object: Response, json_response: bool = True):
        self._response = response_object
        # Should be json or text
        self._json_response = json_response
        if not self.response.ok:
            message = (
                f"The request returned with a {self.response.status_code} {self.reason}"
            )
            if self.response.status_code == 400:
                message += f" -- {self.body}"
            raise HTTPError(message)

    def __call__(self):
        return self.response

    @property
    def json_response(self):
        return self._json_response

    @property
    def response(self):
        return self._response

    @property
    def ok(self):
        return self().ok

    @property
    def status_code(self):
        return self().status_code

    @property
    def reason(self):
        return self().reason

    @property
    def body(self):
        if self._json_response:
            return self().json()
        else:
            return self().text

    @response.setter
    def response(self, new_response):
        if type(new_response) is not Response:
            raise ValueError("response must be a Response object")
        self._response = new_response

    def get_attribute_entries(self, name):
        # Used for any nested data in the response body
        if not self._json_response:
            raise ValueError("get_attribute_entries needs a response of type json")
        body = self.body
        if type(body) is list:
            return [row.get(name) for row in self().json()]
        elif type(body) is dict:
            return body.get(name)


class MetricsResponse(BaseResponse):
    def __init__(self, response_object: Response = None):
        super(MetricsResponse, self).__init__(response_object, False)


class VersionResponse(BaseResponse):
    def __init__(self, response_object: Response = None):
        super(VersionResponse, self).__init__(response_object, True)

    @property
    def app_version(self):
        return self.get_attribute_entries("app_version")


class DiagnosticResponse(BaseResponse):
    def __init__(self, response_object: Response = None):
        super(DiagnosticResponse, self).__init__(response_object, True)

    @property
    def get_platform(self):
        return self.get_attribute_entries("platform")

    @property
    def get_cpu_count(self):
        return self.get_attribute_entries("cpu_count")

    @property
    def get_container_version(self):
        return self.get_attribute_entries("container_version")

    @property
    def get_cpu_name(self):
        return self.get_attribute_entries("cpu_name")

    @property
    def get_gpu_info(self):
        return self.get_attribute_entries("gpu_info")


class DemiTextResponse(BaseResponse):
    def __init__(self, response_object: Response = None):
        super(DemiTextResponse, self).__init__(response_object, True)

    @property
    def processed_text(self):
        return self.get_attribute_entries("processed_text")

    @property
    def entities(self):
        return self.get_attribute_entries("entities")

    @property
    def entities_present(self):
        return self.get_attribute_entries("entities_present")

    @property
    def best_labels(self):
        if type(self.body) == dict:
            best_labels = [entity["best_label"] for entity in self.entities]
        else:
            best_labels = [
                attr["best_label"] for entity in self.entities for attr in entity
            ]
        return best_labels

    def get_reidentify_entities(self):
        if type(self.body) == dict:
            entities = [
                Entity(entity["processed_text"], entity["text"])
                for entity in self.entities
            ]
        else:
            entities = [
                Entity(attr["processed_text"], attr["text"])
                for entity in self.entities
                for attr in entity
            ]
        return entities

    def get_reidentify_request(self):
        entities = self.get_reidentify_entities()
        return ReidentifyTextRequest(self.processed_text, entities)


class NerTextResponse(BaseResponse):
    def __init__(self, response_object: Response = None):
        super(NerTextResponse, self).__init__(response_object, True)

    @property
    def entities(self):
        return self.get_attribute_entries("entities")

    @property
    def entities_present(self):
        return self.get_attribute_entries("entities_present")

    @property
    def characters_processed(self):
        return self.get_attribute_entries("characters_processed")

    @property
    def languages_detected(self):
        return self.get_attribute_entries("languages_detected")


class AnalyzeTextResponse(NerTextResponse):
    def __init__(self, response_object: Response = None):
        super(AnalyzeTextResponse, self).__init__(response_object)

    @property
    def analysis_result(self):
        return self.get_attribute_entries("analysis_result")


class TextResponse(DemiTextResponse):
    def __init__(self, response_object: Response = None):
        super(TextResponse, self).__init__(response_object)

    @property
    def characters_processed(self):
        return self.get_attribute_entries("characters_processed")

    @property
    def languages_detected(self):
        return self.get_attribute_entries("languages_detected")


class FilesUriResponse(DemiTextResponse):
    def __init__(self, response_object: Response = None):
        super(FilesUriResponse, self).__init__(response_object)

    @property
    def result_uri(self):
        return self.get_attribute_entries("result_uri")


class FilesBase64Response(DemiTextResponse):
    def __init__(self, response_object: Response = None):
        super(FilesBase64Response, self).__init__(response_object)

    @property
    def processed_file(self):
        return self.get_attribute_entries("processed_file")


class BleepResponse(BaseResponse):
    def __init__(self, response_object: Response = None):
        super(BleepResponse, self).__init__(response_object, True)

    @property
    def bleeped_file(self):
        return self.get_attribute_entries("bleeped_file")


class ReidentifyTextResponse(BaseResponse):
    def __init__(self, response_object: Response = None):
        super(ReidentifyTextResponse, self).__init__(response_object, True)



import logging


class PAIURIs:
    def __init__(self, url=None, scheme=None, host=None, port=None, **kwargs):
        if url:
            self._pai_uri = url
        elif scheme and host:
            self.valid_schemes = ["http", "https"]
            scheme = scheme.split("://")[0]
            if scheme not in self.valid_schemes:
                raise ValueError(
                    f"Scheme must be one of the following: {', '.join(self.valid_schemes)}"
                )
            port = f":{port}" if port else ""
            self._pai_uri = f"{scheme}://{host}{port}"
        else:
            raise ValueError(
                "PAIClient needs either a url, or a scheme and host to initialize. You can find more information on which url to use here: https://docs.private-ai.com/thin-client/"
            )

    @property
    def pai_uri(self):
        return self._pai_uri

    @property
    def bleep(self):
        return self._create_uri(self.pai_uri, "bleep")

    @property
    def health(self):
        return self._create_uri(self.pai_uri, "healthz")

    @property
    def metrics(self):
        return self._create_uri(self.pai_uri, "metrics")

    @property
    def diagnostics(self):
        return self._create_uri(self.pai_uri, "diagnostics")

    @property
    def process_text(self):
        return self._create_uri(self.pai_uri, "process", "text")

    @property
    def process_files_uri(self):
        return self._create_uri(self.pai_uri, "process", "files", "uri")

    @property
    def reidentify_text(self):
        return self._create_uri(self.pai_uri, "process", "text", "reidentify")

    @property
    def process_files_base64(self):
        return self._create_uri(self.pai_uri, "process", "files", "base64")

    @property
    def ner_text(self):
        return self._create_uri(self.pai_uri, "ner", "text")

    @property
    def analyze_text(self):
        return self._create_uri(self.pai_uri, "analyze", "text")

    @property
    def version(self):
        return self._create_uri(self.pai_uri, "")

    def _create_uri(self, *args):
        return "/".join([x.strip("/") for x in args])



import inspect
from typing import List, Optional, Union


class BaseRequestObject:
    def to_dict(self):
        dict_obj = dict()
        for key, value in self.__dict__.items():
            if value in [None, [], {}]:
                continue
            name = key if key[0] != "_" else key[1:]
            if self._issubclass(value):
                dict_obj[name] = value.to_dict()
            elif type(value) is list:
                dict_obj[name] = [
                    row.to_dict() if self._issubclass(row) else row for row in value
                ]
            elif not key.startswith("__") and not callable(key):
                dict_obj[name] = value
        return dict_obj

    def _issubclass(self, obj):
        return inspect.isclass(type(obj)) and issubclass(type(obj), BaseRequestObject)

    def __call__(self):
        return self.to_dict()

    @classmethod
    def _fromdict(cls, values: dict):
        return cls(**values)


class AudioOptions(BaseRequestObject):
    default_bleep_start_padding: float = 0.5
    default_bleep_end_padding: float = 0.5
    default_bleep_frequency: Optional[int] = None
    default_bleep_gain: Optional[int] = None

    def __init__(
        self,
        bleep_start_padding: float = default_bleep_start_padding,
        bleep_end_padding: float = default_bleep_end_padding,
        bleep_frequency: Optional[int] = default_bleep_frequency,
        bleep_gain: Optional[int] = default_bleep_gain,
    ):
        if self._bleep_start_padding_validator(bleep_start_padding):
            self._bleep_start_padding = bleep_start_padding
        if self._bleep_end_padding_validator(bleep_end_padding):
            self._bleep_end_padding = bleep_end_padding
        if self._bleep_frequency_validator(bleep_frequency):
            self._bleep_frequency = bleep_frequency
        if self._bleep_gain_validator(bleep_gain):
            self._bleep_gain = bleep_gain

    @property
    def bleep_start_padding(self):
        return self._bleep_start_padding

    @property
    def bleep_end_padding(self):
        return self._bleep_end_padding

    @property
    def bleep_frequency(self):
        return self._bleep_frequency

    @property
    def bleep_gain(self):
        return self._bleep_gain

    @bleep_start_padding.setter
    def bleep_start_padding(self, var):
        if self._bleep_start_padding_validator(var):
            self._bleep_start_padding = var

    @bleep_end_padding.setter
    def bleep_end_padding(self, var):
        if self._bleep_end_padding_validator(var):
            self._bleep_end_padding = var

    @bleep_frequency.setter
    def bleep_frequency(self, var):
        if self._bleep_frequency_validator(var):
            self._bleep_frequency = var

    @bleep_gain.setter
    def bleep_gain(self, var):
        if self._bleep_gain_validator(var):
            self._bleep_gain = var

    def _bleep_start_padding_validator(self, var):
        if type(var) is not float:
            raise ValueError(
                f"AudioOptions.bleep_start_padding must be of type float, but got {type(var)}"
            )
        if var < 0:
            raise ValueError("AudioOptions.bleep_start_padding must be positive")
        return True

    def _bleep_end_padding_validator(self, var):
        if type(var) is not float:
            raise ValueError(
                f"AudioOptions.bleep_end_padding must be of type float, but got {type(var)}"
            )
        if var < 0:
            raise ValueError("AudioOptions.bleep_end_padding must be positive")
        return True

    def _bleep_frequency_validator(self, var):
        if type(var) is not int and var is not None:
            raise ValueError(
                f"AudioOptions.bleep_frequency must be of type int or None, but got {type(var)}"
            )
        return True

    def _bleep_gain_validator(self, var):
        if type(var) is not int and var is not None:
            raise ValueError(
                f"AudioOptions.bleep_gain must be of type int or None, but got {type(var)}"
            )
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            return cls._fromdict(values)
        except TypeError:
            raise TypeError(
                "ProcessedText can only accept the values 'bleep_start_padding', 'bleep_end_padding', 'bleep_frequency', and 'bleep_gain'"
            )


class ImageOptions(BaseRequestObject):
    default_masking_method: str = "blur"
    default_palette: bool = False
    VALID_MASK_MODES = ["blur", "blackbox"]

    def __init__(
        self,
        masking_method: str = default_masking_method,
        palette: Optional[bool] = default_palette,
    ):
        if self._masking_method_validator(masking_method):
            self._masking_method = masking_method
        if self._palette_validator(palette):
            self._palette = palette

    @property
    def masking_method(self):
        return self._masking_method

    @property
    def palette(self):
        return self._palette

    @masking_method.setter
    def masking_method(self, var):
        if self._masking_method_validator(var):
            self._masking_method = var

    @palette.setter
    def palette(self, var):
        if self._palette_validator(var):
            self._palette = var

    def _masking_method_validator(self, var):
        if var not in self.VALID_MASK_MODES:
            raise ValueError(
                f"ImageOptions.masking_method must be one of {self.VALID_MASK_MODES}, but got {var}"
            )
        return True

    def _palette_validator(self, var):
        if not isinstance(var, bool):
            raise ValueError(
                f"ImageOptions.palette must be of type bool, but got {type(var)}"
            )
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            return cls._fromdict(values)
        except TypeError:
            raise TypeError(
                "ImageOptions can only accept the values 'masking_method' and 'palette'"
            )


class Entity(BaseRequestObject):
    def __init__(self, processed_text: str, text: str):
        if self._processed_text_validator(processed_text):
            self._processed_text = processed_text
        if self._text_validator(text):
            self._text = text

    @property
    def processed_text(self):
        return self._processed_text

    @property
    def text(self):
        return self._text

    @processed_text.setter
    def processed_text(self, var):
        if self._processed_text_validator(var):
            self._processed_text = var

    @text.setter
    def text(self, var):
        if self._text_validator(var):
            self._text = var

    def _processed_text_validator(self, var):
        if type(var) is not str:
            raise TypeError(
                f"{var} is not valid. Entity.processed_text must be of type string"
            )
        return True

    def _text_validator(self, var):
        if type(var) is not str:
            raise TypeError(f"{var} is not valid. Entity.text must be of type string")
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            return cls._fromdict(values)
        except TypeError:
            raise TypeError(
                "Entity can only accept the values 'processed_text' and 'text'"
            )


class EntityTypeSelector(BaseRequestObject):
    valid_types = ["DISABLE", "ENABLE"]

    def __init__(self, type: str, value: List[str] = []):
        if self._type_validator(type):
            self._type = type
        if self._value_validator(value):
            self.value = value

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, var):
        if self._type_validator(var):
            self._type = var

    def _type_validator(self, var):
        if var not in self.valid_types:
            raise ValueError(
                f"'{var}' is not valid. EntityTypeSelector.type can only be one of the following: {', '.join(self.valid_types)}"
            )
        return True

    def _value_validator(self, var):
        if type(var) is not list:
            raise TypeError("EntityTypeSelector.value must be of type list")
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            return cls._fromdict(values)
        except TypeError:
            raise TypeError(
                "EntityTypeSelector can only accept the values 'type' and 'value'"
            )


class File(BaseRequestObject):
    valid_content_types = [
        "application/json",
        "application/msword",
        "application/vnd.ms-powerpoint",
        "application/pdf",
        "application/vnd.ms-excel",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        "application/vnd.openxmlformats-officedocument.presentationml.presentation",
        "application/xml",
        "message/rfc822",
        "text/csv",
        "text/plain",
        "image/jpeg",
        "image/jpg",
        "image/tif",
        "image/tiff",
        "image/png",
        "image/bmp",
        "application/dicom",
        "audio/mp4",
        "audio/m4a",
        "audio/mp4a-latm",
        "audio/mp3",
        "audio/mpeg",
        "audio/wav",
        "audio/x-wav",
    ]

    def __init__(self, data: str, content_type: str):
        if self._data_validator(data):
            self._data = data
        if self._content_type_validator(content_type):
            self._content_type = content_type

    @property
    def data(self):
        return self._data

    @property
    def content_type(self):
        return self._content_type

    @data.setter
    def data(self, var):
        if self._data_validator(var):
            self._data = var

    @content_type.setter
    def content_type(self, var):
        if self._content_type_validator(var):
            self._content_type = var

    def _data_validator(self, var):
        if type(var) is not str:
            raise TypeError("data must be string-type")
        return True

    def _content_type_validator(self, var):
        if var not in self.valid_content_types:
            raise ValueError(
                f"{var} is not valid. File.content_type can only be one of the following: {', '.join(self.valid_content_types)}"
            )
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            return cls._fromdict(values)
        except TypeError:
            raise TypeError("File can only accept the values 'data' and 'content_type'")


class FilterSelector(BaseRequestObject):
    valid_types = ["ALLOW", "BLOCK", "ALLOW_TEXT"]
    default_threshold = 1

    def __init__(
        self,
        type: str,
        pattern: str,
        entity_type: str = None,
        threshold: Union[int, float] = default_threshold,
    ):
        if self._type_validator(type):
            self._type = type
        if self._pattern_validator(pattern):
            self._pattern = pattern
        if self.type == "BLOCK":
            if self._entity_type_validator(entity_type):
                self._entity_type = entity_type
            if self._threshold_validator(threshold):
                self._threshold = threshold

    @property
    def type(self):
        return self._type

    @property
    def pattern(self):
        return self._pattern

    @property
    def entity_type(self):
        if self.type != "BLOCK":
            raise AttributeError(
                f"FilterSelector of type {self.type} does not contain entity_type"
            )
        return self._entity_type

    @property
    def threshold(self):
        if self.type != "BLOCK":
            raise AttributeError(
                f"FilterSelector of type {self.type} does not contain threshold"
            )
        return self._threshold

    @type.setter
    def type(self, var):
        if self._type_validator(var):
            self._type = var

    @pattern.setter
    def pattern(self, var):
        if self._pattern_validator(var):
            self._pattern = var

    @entity_type.setter
    def entity_type(self, var):
        if self._entity_type_validator(var):
            self._entity_type = var

    @threshold.setter
    def threshold(self, var):
        if self._threshold_validator(var):
            self._threshold = var

    def _type_validator(self, var):
        if var not in self.valid_types:
            raise ValueError(
                f"{var} is not valid. FilterSelector.type can only be one of the following: {', '.join(self.valid_types)}"
            )
        return True

    def _pattern_validator(self, var):
        if type(var) is not str:
            raise TypeError("FilterSelector.pattern must be of type string")
        return True

    def _entity_type_validator(self, var):
        if type(var) is not str:
            raise TypeError("FilterSelector.entity_type must be of type string")
        return True

    def _threshold_validator(self, var):
        if var < 0:
            raise TypeError("FilterSelector.threshold must be greater than 0")
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            return cls._fromdict(values)
        except TypeError:
            raise TypeError(
                "FilterSelector can only accept the values 'type' and 'pattern'"
            )


class PDFOptions(BaseRequestObject):
    default_density = 200
    default_max_resolution = 3000
    default_enable_pdf_text_layer: bool = True

    def __init__(
        self,
        density: int = default_density,
        max_resolution: int = default_max_resolution,
        enable_pdf_text_layer: bool = default_enable_pdf_text_layer,
    ):
        self._density = density
        self._max_resolution = max_resolution
        self._enable_pdf_text_layer = enable_pdf_text_layer

    @property
    def density(self):
        return self._density

    @property
    def max_resolution(self):
        return self._max_resolution

    @property
    def enable_pdf_text_layer(self):
        return self._enable_pdf_text_layer

    @density.setter
    def density(self, var):
        if self._density_validator(var):
            self._density = var

    @max_resolution.setter
    def max_resolution(self, var):
        if self._max_resolution_validator(var):
            self._max_resolution = var

    @enable_pdf_text_layer.setter
    def enable_pdf_text_layer(self, var):
        if self._enable_pdf_text_layer_validator(var):
            self._enable_pdf_text_layer = var

    def _density_validator(self, var):
        if type(var) is not int:
            raise ValueError("PDFOptions.density must be of type int and >0")
        return True

    def _max_resolution_validator(self, var):
        if type(var) is not int:
            raise ValueError("PDFOptions.max_resolution must be of type int and >0")
        return True

    def _enable_pdf_text_layer_validator(self, var):
        if type(var) is not bool:
            raise ValueError("PDFOptions.enable_pdf_text_layer must be of type bool")
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            return cls._fromdict(values)
        except TypeError:
            raise TypeError(
                "PDFOptions can only accept 'density', 'max_resolution' and 'enable_pdf_text_layer'"
            )


class OCROptions(BaseRequestObject):
    default_ocr_system = "paddleocr"
    VALID_OCR_SYSTEM = [
        "azure_computer_vision",
        "azure_doc_intelligence",
        "hybrid",
        "paddleocr",
    ]

    def __init__(self, ocr_system: str = default_ocr_system):
        self._ocr_system = ocr_system

    @property
    def ocr_system(self):
        return self._ocr_system

    @ocr_system.setter
    def ocr_system(self, var):
        if self._ocr_system_validator(var):
            self._ocr_system = var

    def _ocr_system_validator(self, var):
        if var not in self.VALID_OCR_SYSTEM:
            raise ValueError(
                f"OCROptions.ocr_system must be one of {','.join(self.VALID_OCR_SYSTEM)}, but got {var}"
            )
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            return cls._fromdict(values)
        except TypeError:
            raise TypeError("OCROptions can only accept 'ocr_system'")


class ObjectEntityTypeSelector(BaseRequestObject):
    valid_types = ["DISABLE", "ENABLE"]
    valid_values = {"FACE", "LICENSE_PLATE", "LOGO", "SIGNATURE"}

    def __init__(self, type: str, value: List[str] = []):
        if self._type_validator(type):
            self._type = type
        if self._value_validator(value):
            self.value = value

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, var):
        if self._type_validator(var):
            self._type = var

    def _type_validator(self, var):
        if var not in self.valid_types:
            raise ValueError(
                f"'{var}' is not valid. ObjectEntityTypeSelector.type can only be one of the following: {', '.join(self.valid_types)}"
            )
        return True

    def _value_validator(self, var):
        if type(var) is not list:
            raise TypeError("ObjectEntityTypeSelector.value must be of type list")
        elif var and not all(value in self.valid_values for value in var):
            raise ValueError(
                f"'{var}' is not valid. ObjectEntityTypeSelector.value can only be one of the following: {', '.join(self.valid_values)}"
            )
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            return cls._fromdict(values)
        except TypeError:
            raise TypeError(
                "ObjectEntityTypeSelector can only accept the values 'type' and 'value'"
            )


class ProcessedMarkerText(BaseRequestObject):
    attributes = ["_pattern"]
    default_pattern = "[UNIQUE_NUMBERED_ENTITY_TYPE]"
    valid_patterns = [
        "BEST_ENTITY_TYPE",
        "ALL_ENTITY_TYPES",
        "UNIQUE_NUMBERED_ENTITY_TYPE",
        "UNIQUE_HASHED_ENTITY_TYPE",
    ]
    default_marker_language = "en"
    valid_marker_languages = ["auto", "en", "fr", "de", "ja", "ko", "nl", "ru", "uk"]
    default_coreference_resolution = "heuristics"
    valid_coreference_resolutions = ["heuristics", "model_prediction", "combined"]

    def __init__(
        self,
        pattern: str = default_pattern,
        marker_language: str = default_marker_language,
        coreference_resolution: str = default_coreference_resolution,
    ):
        for attribute in (
            ProcessedMaskText.attributes + ProcessedSyntheticText.attributes
        ):
            delattr(self, attribute) if hasattr(self, attribute) else False
        self._type = "MARKER"
        if self._pattern_validator(pattern):
            self._pattern = pattern
        if self._marker_language_validator(marker_language):
            self._marker_language = marker_language
        if self._coreference_resolution_validator(coreference_resolution):
            self._coreference_resolution = coreference_resolution

    @property
    def pattern(self):
        return self._pattern

    @property
    def marker_language(self):
        return self._marker_language

    @property
    def coreference_resolution(self):
        return self._coreference_resolution

    @pattern.setter
    def pattern(self, var):
        if self._pattern_validator(var):
            self._pattern = var

    @marker_language.setter
    def marker_language(self, var):
        if self._marker_language_validator(var):
            self._marker_language = var

    @coreference_resolution.setter
    def coreference_resolution(self, var):
        if self._coreference_resolution_validator(var):
            self._coreference_resolution = var

    def _pattern_validator(self, var):
        if var not in self.valid_patterns and var[1:-1] not in self.valid_patterns:
            raise ValueError(
                f"{var} is not valid. ProcessedText.pattern can only be one of the following: {', '.join(self.valid_patterns)}"
            )
        return True

    def _marker_language_validator(self, var):
        if var not in self.valid_marker_languages:
            raise ValueError(
                f"{var} is not valid. ProcessedText.marker_language can only be one of the following: {', '.join(self.valid_marker_languages)}"
            )
        return True

    def _coreference_resolution_validator(self, var):
        if var not in self.valid_coreference_resolutions:
            raise ValueError(
                f"{var} is not valid. ProcessedText.coreference_resolution can only be one of the following: {', '.join(self.valid_coreference_resolutions)}"
            )
        return True


class ProcessedMaskText(BaseRequestObject):
    attributes = ["_mask_character"]

    def __init__(self, mask_character: str = "#"):
        for attribute in (
            ProcessedMarkerText.attributes + ProcessedSyntheticText.attributes
        ):
            delattr(self, attribute) if hasattr(self, attribute) else False
        if self._mask_character_validator(mask_character):
            self._mask_character = mask_character
        self._type = "MASK"

    @property
    def mask_character(self):
        return self._mask_character

    @mask_character.setter
    def mask_character(self, var):
        if self._mask_character_validator(var):
            self._mask_character = var

    def _mask_character_validator(self, var):
        if len(var) != 1:
            raise ValueError(
                f"mask_character must have only one character. {var} has {len(var)} characters."
            )
        return True


class ProcessedSyntheticText(BaseRequestObject):
    attributes = ["_synthetic_entity_accuracy", "_preserve_relationships"]
    valid_synthetic_accuracy_values = ["standard", "standard_multilingual"]

    def __init__(
        self,
        synthetic_entity_accuracy: str = "standard",
        preserve_relationships: bool = True,
    ):
        for attribute in ProcessedMarkerText.attributes + ProcessedMaskText.attributes:
            delattr(self, attribute) if hasattr(self, attribute) else False
        self._type = "SYNTHETIC"
        if self._synthetic_entity_accuracy_validator(synthetic_entity_accuracy):
            self._synthetic_entity_accuracy = synthetic_entity_accuracy
        self._preserve_relationships = preserve_relationships
        self._preserve_relationships = True

    @property
    def synthetic_entity_accuracy(self):
        return self._synthetic_entity_accuracy

    @synthetic_entity_accuracy.setter
    def synthetic_entity_accuracy(self, var):
        if self._synthetic_entity_accuracy_validator(var):
            self._synthetic_entity_accuracy = var

    @property
    def preserve_relationships(self):
        return self._preserve_relationships

    @preserve_relationships.setter
    def preserve_relationships(self, var):
        self._preserve_relationships = var

    def _synthetic_entity_accuracy_validator(self, var):
        if var not in self.valid_synthetic_accuracy_values:
            raise ValueError(
                f"{var} is not valid. Synthetic Entity Accuracy can only accept values {', '.join(self.valid_synthetic_accuracy_values)}"
            )


class ProcessedText(ProcessedMarkerText, ProcessedMaskText, ProcessedSyntheticText):
    default_type = "MARKER"
    valid_types = ["MARKER", "MASK", "SYNTHETIC"]

    def __init__(self, type: str = default_type, **kwargs):
        if type == "MARKER":
            ProcessedMarkerText.__init__(self, **kwargs)
        elif type == "MASK":
            ProcessedMaskText.__init__(self, **kwargs)
        elif type == "SYNTHETIC":
            ProcessedSyntheticText.__init__(self, **kwargs)
        else:
            raise ValueError(
                f"{type} is not valid. ProcessedText.type can only be one of the following: {', '.join(self.valid_types)}"
            )

    @classmethod
    def fromdict(cls, values: dict):
        try:
            return cls._fromdict(values)
        except TypeError:
            raise TypeError(
                "ProcessedText can only accept the values 'type' and 'pattern'"
            )

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, var):
        if var in self.valid_types and var != self._type:
            self.__init__(var)


class Timestamp(BaseRequestObject):
    def __init__(self, start, end):
        if self._start_validator(start):
            self._start = start
        if self._end_validator(end):
            self._end = end

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @start.setter
    def start(self, var):
        if self._start_validator(var):
            self._start = var

    @end.setter
    def end(self, var):
        if self._end_validator(var):
            self._end = var

    def _start_validator(self, var):
        if type(var) is not float:
            raise ValueError("Timestamp.start must be of type float")
        return True

    def _end_validator(self, var):
        if type(var) is not float:
            raise ValueError("Timestamp.end must be of type float")
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            return cls._fromdict(values)
        except TypeError:
            raise TypeError("Timestamp can only accept the values 'start' and 'end'")


class EntityDetection(BaseRequestObject):
    default_accuracy = "high_automatic"
    default_return_entity = True
    valid_accuracies = [
        "standard",
        "standard_high",
        "standard_high_multilingual",
        "high",
        "high_multilingual",
        "standard_high_automatic",
        "high_automatic",
    ]
    default_enable_non_max_suppression = False

    def __init__(
        self,
        accuracy: str = default_accuracy,
        entity_types: List[EntityTypeSelector] = [],
        filter: List[FilterSelector] = [],
        return_entity: bool = default_return_entity,
        enable_non_max_suppression: bool = default_enable_non_max_suppression,
    ):
        if self._accuracy_validator(accuracy):
            self._accuracy = accuracy
        if self._entity_types_validator(entity_types):
            self.entity_types = entity_types
        if self._filter_validator(filter):
            self.filter = filter
        if self._return_entity_validator(return_entity):
            self._return_entity = return_entity
        if self._enable_non_max_suppression_validator(enable_non_max_suppression):
            self._enable_non_max_suppression = enable_non_max_suppression

    @property
    def accuracy(self):
        return self._accuracy

    @property
    def return_entity(self):
        return self._return_entity

    @accuracy.setter
    def accuracy(self, var):
        if self._accuracy_validator(var):
            self._accuracy = var

    @return_entity.setter
    def return_entity(self, var):
        if self._return_entity_validator(var):
            self._return_entity = var

    def _accuracy_validator(self, var):
        if var not in self.valid_accuracies:
            raise ValueError(
                f"{var} is not valid. EntityDetection.accuracy can only be one of the following: {', '.join(self.valid_accuracies)}"
            )
        return True

    def _entity_types_validator(self, var):
        if type(var) is not list:
            raise TypeError(
                f"{var} is not valid. EntityDetection.entity_types can only be a list"
            )
        elif var and not all(isinstance(row, EntityTypeSelector) for row in var):
            raise ValueError(
                "EntityDetection.entity_types can only contain EntityTypeSelector objects"
            )
        return True

    def _filter_validator(self, var):
        if type(var) is not list:
            raise ValueError(
                f"{var} is not valid. EntityDetection.filter can only be a list"
            )
        elif var and not all(isinstance(x, FilterSelector) for x in var):
            raise ValueError(
                "EntityDetection.filter can only contain FilterSelector objects"
            )
        return True

    def _return_entity_validator(self, var):
        if type(var) is not bool:
            raise ValueError("EntityDetection.return_entity must be of type bool")
        return True

    def _enable_non_max_suppression_validator(self, var):
        if type(var) is not bool:
            raise ValueError(
                "EntityDetection.enable_non_max_suppression must be of type bool"
            )
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            initializer_dict = {}
            for key, value in values.items():
                if key == "entity_types":
                    initializer_dict[key] = [
                        EntityTypeSelector.fromdict(row) for row in value
                    ]
                elif key == "filter":
                    initializer_dict[key] = [
                        FilterSelector.fromdict(row) for row in value
                    ]
                else:
                    initializer_dict[key] = value
            return cls._fromdict(initializer_dict)
        except TypeError:
            raise TypeError(
                "EntityDetection can only accept the values 'accuracy', 'entity_types', 'filter' and 'return_entity'"
            )


class ObjectEntityDetection(BaseRequestObject):
    def __init__(self, object_entity_types: List[ObjectEntityTypeSelector] = []):
        if self._object_entity_types_validator(object_entity_types):
            self.object_entity_types = object_entity_types

    def _object_entity_types_validator(self, var):
        if type(var) is not list:
            raise TypeError(
                f"{var} is not valid. ObjectEntityDetection.object_entity_types can only be a list"
            )
        elif var and not all(isinstance(row, ObjectEntityTypeSelector) for row in var):
            raise ValueError(
                "ObjectEntityDetection.object_entity_types can only contain ObjectEntityTypeSelector objects"
            )
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            initializer_dict = {}
            for key, value in values.items():
                if key == "object_entity_types":
                    initializer_dict[key] = [
                        ObjectEntityTypeSelector.fromdict(row) for row in value
                    ]
                else:
                    initializer_dict[key] = value
            return cls._fromdict(initializer_dict)
        except TypeError:
            raise TypeError(
                "ObjectEntityDetection can only accept the value 'object_entity_types'"
            )


class RelationDetection(BaseRequestObject):
    default_coreference_resolution = None
    valid_coreference_resolutions = ["heuristics", "model_prediction", "combined"]

    def __init__(
        self, coreference_resolution: Optional[str] = default_coreference_resolution
    ):
        if self._coreference_resolution_validator:
            self._coreference_resolution = coreference_resolution

    @property
    def coreference_resolution(self):
        return self._coreference_resolution

    @coreference_resolution.setter
    def coreference_resolution(self, var):
        if self._coreference_resolution_validator(var):
            self._coreference_resolution = var

    def _coreference_resolution_validator(self, var):
        if var is None:
            return True
        if var not in self.valid_coreference_resolutions:
            raise ValueError(
                f"{var} is not valid. RelationDetection.coreference_resolution can only be one of the following: {', '.join(self.valid_coreference_resolutions)} or None"
            )
        return True

    @classmethod
    def fromdict(cls, values: dict):
        try:
            initializer_dict = {}
            for key, value in values.items():
                initializer_dict[key] = value
            return cls._fromdict(initializer_dict)
        except TypeError:
            raise TypeError(
                "RelationDetection can only accept the value 'coreference_resolution'"
            )


class ProcessTextRequest(BaseRequestObject):
    def __init__(
        self,
        text: List[str],
        link_batch: Optional[bool] = None,
        entity_detection: Optional[EntityDetection] = None,
        processed_text: Optional[ProcessedText] = None,
        project_id: Optional[str] = None,
    ):
        self.text = text
        self.link_batch = link_batch
        self.entity_detection = entity_detection
        self.processed_text = processed_text
        self.project_id = project_id

    @classmethod
    def fromdict(cls, values: dict):
        try:
            initializer_dict = {}
            for key, value in values.items():
                if key == "entity_detection":
                    initializer_dict[key] = EntityDetection.fromdict(value)
                elif key == "processed_text":
                    initializer_dict[key] = ProcessedText.fromdict(value)
                else:
                    initializer_dict[key] = value
            return cls._fromdict(initializer_dict)
        except TypeError:
            raise TypeError(
                "ProcessTextRequest can only accept the values 'text', 'link_batch', 'entity_detection' and 'process_text'"
            )


class NerTextRequest(BaseRequestObject):
    def __init__(
        self,
        text: List[str],
        link_batch: Optional[bool] = None,
        entity_detection: Optional[EntityDetection] = None,
        project_id: Optional[str] = None,
    ):
        self.text = text
        self.link_batch = link_batch
        self.entity_detection = entity_detection
        self.project_id = project_id

    @classmethod
    def fromdict(cls, values: dict):
        try:
            initializer_dict = {}
            for key, value in values.items():
                if key == "entity_detection":
                    initializer_dict[key] = EntityDetection.fromdict(value)
                else:
                    initializer_dict[key] = value
            return cls._fromdict(initializer_dict)
        except TypeError:
            raise TypeError(
                "NerTextRequest can only accept the values 'text', 'link_batch' and 'entity_detection'"
            )


class AnalyzeTextRequest(BaseRequestObject):
    def __init__(
        self,
        text: List[str],
        locale: Optional[str] = None,
        link_batch: Optional[bool] = None,
        entity_detection: Optional[EntityDetection] = None,
        project_id: Optional[str] = None,
        relation_detection: Optional[RelationDetection] = None,
    ):
        self.text = text
        self.locale = locale
        self.link_batch = link_batch
        self.entity_detection = entity_detection
        self.project_id = project_id
        self.relation_detection = relation_detection

    @classmethod
    def fromdict(cls, values: dict):
        try:
            initializer_dict = {}
            for key, value in values.items():
                if key == "entity_detection":
                    initializer_dict[key] = EntityDetection.fromdict(value)
                elif key == "relation_detection":
                    initializer_dict[key] = RelationDetection.fromdict(value)
                else:
                    initializer_dict[key] = value

            return cls._fromdict(initializer_dict)
        except TypeError:
            raise TypeError(
                "AnalyzeTextRequest can only accept the values 'text', 'locale', 'link_batch', 'entity_detection' and 'relation_detection'"
            )


class ProcessFileUriRequest(BaseRequestObject):
    def __init__(
        self,
        uri: str,
        entity_detection: Optional[EntityDetection] = None,
        object_entity_detection: Optional[ObjectEntityDetection] = None,
        pdf_options: Optional[PDFOptions] = None,
        audio_options: Optional[AudioOptions] = None,
        image_options: Optional[ImageOptions] = None,
        project_id: Optional[str] = None,
        ocr_options: Optional[OCROptions] = None,
    ):
        self.uri = uri
        self.entity_detection = entity_detection
        self.object_entity_detection = object_entity_detection
        self.pdf_options = pdf_options
        self.audio_options = audio_options
        self.image_options = image_options
        self.project_id = project_id
        self.ocr_options = ocr_options

    @classmethod
    def fromdict(cls, values: dict):
        try:
            initializer_dict = {}
            for key, value in values.items():
                if key == "entity_detection":
                    initializer_dict[key] = EntityDetection.fromdict(value)
                elif key == "object_entity_detection":
                    initializer_dict[key] = ObjectEntityDetection.fromdict(value)
                elif key == "pdf_options":
                    initializer_dict[key] = PDFOptions.fromdict(value)
                elif key == "audio_options":
                    initializer_dict[key] = AudioOptions.fromdict(value)
                elif key == "image_options":
                    initializer_dict[key] = ImageOptions.fromdict(value)
                elif key == "ocr_options":
                    initializer_dict[key] = OCROptions.fromdict(value)
                else:
                    initializer_dict[key] = value
            return cls._fromdict(initializer_dict)
        except TypeError:
            raise TypeError(
                "ProcessFileUriRequest can only accept the values 'uri', 'entity_detection', 'object_entity_detection', 'pdf_options', 'audio_options', 'image_options' and 'ocr_options'"
            )


class ProcessFileBase64Request(BaseRequestObject):
    def __init__(
        self,
        file: File,
        entity_detection: Optional[EntityDetection] = None,
        object_entity_detection: Optional[ObjectEntityDetection] = None,
        pdf_options: Optional[PDFOptions] = None,
        audio_options: Optional[AudioOptions] = None,
        image_options: Optional[ImageOptions] = None,
        project_id: Optional[str] = None,
        ocr_options: Optional[OCROptions] = None,
    ):
        self.file = file
        self.entity_detection = entity_detection
        self.object_entity_detection = object_entity_detection
        self.pdf_options = pdf_options
        self.audio_options = audio_options
        self.image_options = image_options
        self.project_id = project_id
        self.ocr_options = ocr_options

    @classmethod
    def fromdict(cls, values: dict):
        try:
            initializer_dict = {}
            for key, value in values.items():
                if key == "file":
                    initializer_dict[key] = File.fromdict(value)
                elif key == "entity_detection":
                    initializer_dict[key] = EntityDetection.fromdict(value)
                elif key == "object_entity_detection":
                    initializer_dict[key] = ObjectEntityDetection.fromdict(value)
                elif key == "pdf_options":
                    initializer_dict[key] = PDFOptions.fromdict(value)
                elif key == "audio_options":
                    initializer_dict[key] = AudioOptions.fromdict(value)
                elif key == "image_options":
                    initializer_dict[key] = ImageOptions.fromdict(value)
                elif key == "ocr_options":
                    initializer_dict[key] = OCROptions.fromdict(value)
                else:
                    initializer_dict[key] = value
            return cls._fromdict(initializer_dict)
        except TypeError:
            raise TypeError(
                "ProcessFileBase64Request can only accept the values 'file', 'entity_detection', 'object_entity_detection', 'pdf_options', 'audio_options', 'image_options' and 'ocr_options'"
            )


class BleepRequest(BaseRequestObject):
    def __init__(
        self,
        file: File,
        timestamps: List,
        bleep_frequency: Optional[int] = None,
        bleep_gain: Optional[int] = None,
    ):
        self.file = file
        self.timestamps = timestamps
        self.bleep_frequency = bleep_frequency
        self.bleep_gain = bleep_gain

    @classmethod
    def fromdict(cls, values: dict):
        try:
            initializer_dict = {}
            for key, value in values.items():
                if key == "file":
                    initializer_dict[key] = File.fromdict(value)
                elif key == "timestamps":
                    initializer_dict[key] = [
                        Timestamp.fromdict(entry) for entry in value
                    ]
                else:
                    initializer_dict[key] = value
            return cls._fromdict(initializer_dict)
        except TypeError:
            raise TypeError(
                "BleepRequest can only accept the values 'file', 'timestamps', 'bleep_frequency', and 'bleep_gain'"
            )


class ReidentifyTextRequest(BaseRequestObject):
    def __init__(
        self,
        processed_text: List[str] = [],
        entities: List[Entity] = [],
        model: str = "",
        reidentify_sensitive_fields: bool = True,
    ):
        self.reidentify_sensitive_fields = reidentify_sensitive_fields
        self.processed_text = processed_text
        self.entities = entities
        self.model = model

    @classmethod
    def fromdict(cls, values: dict):
        try:
            initializer_dict = {}
            for key, value in values.items():
                if key == "entities":
                    initializer_dict[key] = [
                        Entity.fromdict(entity) for entity in values[key]
                    ]
                else:
                    initializer_dict[key] = value
            return cls._fromdict(initializer_dict)
        except TypeError:
            raise TypeError(
                "ReidentifyTextRequest can only accept the values 'processed_text', 'entities', 'model' and 'reidentify_sensitive_fields'"
            )



from .components import request_objects as req_objects


class request_objects:
    bleep_obj = req_objects.BleepRequest
    file_uri_obj = req_objects.ProcessFileUriRequest
    file_base64_obj = req_objects.ProcessFileBase64Request
    process_text_obj = req_objects.ProcessTextRequest
    ner_text_obj = req_objects.NerTextRequest
    analyze_text_obj = req_objects.AnalyzeTextRequest
    audio_options_obj = req_objects.AudioOptions
    image_options_obj = req_objects.ImageOptions
    entity = req_objects.Entity
    entity_type_selector_obj = req_objects.EntityTypeSelector
    object_entity_type_selector_obj = req_objects.ObjectEntityTypeSelector
    entity_detection_obj = req_objects.EntityDetection
    object_entity_detection_obj = req_objects.ObjectEntityDetection
    file_obj = req_objects.File
    filter_selector_obj = req_objects.FilterSelector
    pdf_options_obj = req_objects.PDFOptions
    ocr_options_obj = req_objects.OCROptions
    processed_text_obj = req_objects.ProcessedText
    reidentify_text_obj = req_objects.ReidentifyTextRequest
    timestamp_obj = req_objects.Timestamp
    relation_detection_obj = req_objects.RelationDetection



import logging
from typing import Union

from .__about__ import __version__
from .components import *


class PAIClient:
    """
    Client used to connect to private-ai's deidentication service
    """

    def __init__(
        self,
        scheme: str = None,
        host: str = None,
        port: str = None,
        url: str = None,
        **kwargs,
    ):
        # Add source url
        self._uris = PAIURIs(url, scheme, host, port)
        self.get = PAIGetRequests(self._uris)
        self.post = PAIPostRequests(self._uris)
        if "api_key" in kwargs.keys():
            self.add_api_key(kwargs["api_key"])
        elif "bearer_token" in kwargs.keys():
            self.add_bearer_token(kwargs["bearer_token"])
        self._container_version = None

    def _add_auth(self, auth_type, auth_val):
        auth_header = {}
        if auth_type == "api_key":
            auth_header = {"x-api-key": auth_val}
        elif auth_type == "bearer_token":
            auth_header = {"Authorization": f"Bearer {auth_val}"}
        else:
            raise ValueError(
                f"{auth_type} is not currently a supported method of authorization"
            )
        for subclass in [self.get, self.post]:
            subclass.headers = {**auth_header, **subclass.base_header}

    def _version_warning(self) -> None:
        # only compare major and minor version numbers
        parsed_container_version = self._container_version.split(".")[:2]
        parsed_client_version = __version__.split(".")[:2]

        if parsed_container_version != parsed_client_version:
            logging.warning(
                f"Version mismatch: privateai_client {__version__} may be incompatible with PAI container "
                f"{self._container_version}. Please install the appropriate client version!"
            )

    def check_version_compatibility(self) -> None:
        if self._container_version is None:
            self.get_version()
        else:
            self._version_warning()

    def add_api_key(self, api_key: str):
        self._add_auth("api_key", api_key)

    def add_bearer_token(self, token: str):
        self._add_auth("bearer_token", token)

    def ping(self):
        """
        Makes a call to the Private-AI service's health endpoint.
        Can be used as a validator to ensure the service is running.
        """
        response = self.get.health()
        if response.status_code != 200:
            logging.warning(f"The Private AI server cannot be reached")
            return False
        return True

    def get_metrics(self):
        """
        Returns information about the Private-AI's server
        """
        self.check_version_compatibility()
        return MetricsResponse(self.get.metrics())

    def get_version(self):
        """
        Returns the version of the container application code
        """
        ret = VersionResponse(self.get.version())
        self._container_version = ret.app_version
        self._version_warning()
        return ret

    def get_diagnostics(self):
        """
        Returns diagnostic information about the Private-AI container host
        """
        self.check_version_compatibility()
        return DiagnosticResponse(self.get.diagnostics())

    def process_text(self, request_object: Union[dict, ProcessTextRequest]):
        """
        Used to deidentify text
        """
        if type(request_object) is ProcessTextRequest:
            self.check_version_compatibility()
            response = TextResponse(self.post.process_text(request_object.to_dict()))
        elif type(request_object) is dict:
            self.check_version_compatibility()
            response = TextResponse(self.post.process_text(request_object))
        else:
            raise ValueError(
                "request_object can only be a dictionary or a ProcessTextRequest object"
            )
        return response

    def reidentify_text(self, request_object: Union[dict, ReidentifyTextRequest]):
        """
        Used to reidentify text
        """
        if type(request_object) is ReidentifyTextRequest:
            self.check_version_compatibility()
            response = ReidentifyTextResponse(
                self.post.reidentify_text(request_object.to_dict())
            )
        elif type(request_object) is dict:
            self.check_version_compatibility()
            response = ReidentifyTextResponse(self.post.reidentify_text(request_object))
        else:
            raise ValueError(
                "request_object can only be a dictionary or a ReidentifyTextRequest object"
            )
        return response

    def process_files_uri(self, request_object: Union[dict, ProcessFileUriRequest]):
        """
        Used to deidentify files by uri
        """
        if type(request_object) is ProcessFileUriRequest:
            self.check_version_compatibility()
            response = FilesUriResponse(
                self.post.process_files_uri(request_object.to_dict())
            )
        elif type(request_object) is dict:
            self.check_version_compatibility()
            response = FilesUriResponse(self.post.process_files_uri(request_object))
        else:
            raise ValueError(
                "request_object can only be a dictionary or a ProcessFileUriRequest object"
            )
        return response

    def process_files_base64(
        self, request_object: Union[dict, ProcessFileBase64Request]
    ):
        """
        Used to deidentify base64 files
        """
        if type(request_object) is ProcessFileBase64Request:
            self.check_version_compatibility()
            response = FilesBase64Response(
                self.post.process_files_base64(request_object.to_dict())
            )
        elif type(request_object) is dict:
            self.check_version_compatibility()
            response = FilesBase64Response(
                self.post.process_files_base64(request_object)
            )
        else:
            raise ValueError(
                "request_object can only be a dictionary or a ProcessFileBase64Request object"
            )
        return response

    def bleep(self, request_object: Union[dict, BleepRequest]):
        """
        Used to deidentify audio files by uri
        """
        if type(request_object) is BleepRequest:
            self.check_version_compatibility()
            response = BleepResponse(self.post.bleep(request_object.to_dict()))
        elif type(request_object) is dict:
            self.check_version_compatibility()
            response = BleepResponse(self.post.bleep(request_object))
        else:
            raise ValueError(
                "request_object can only be a dictionary or a BleepRequest object"
            )
        return response

    def ner_text(self, request_object: Union[dict, NerTextRequest]):
        """
        Used to deidentify text
        """
        if type(request_object) is NerTextRequest:
            self.check_version_compatibility()
            response = NerTextResponse(self.post.ner_text(request_object.to_dict()))
        elif type(request_object) is dict:
            self.check_version_compatibility()
            response = NerTextResponse(self.post.ner_text(request_object))
        else:
            raise ValueError(
                "request_object can only be a dictionary or a NerTextRequest object"
            )
        return response

    def analyze_text(self, request_object: Union[dict, AnalyzeTextRequest]):
        """
        Used to analyze text
        """
        if type(request_object) is AnalyzeTextRequest:
            self.check_version_compatibility()
            response = AnalyzeTextResponse(
                self.post.analyze_text(request_object.to_dict())
            )
        elif type(request_object) is dict:
            self.check_version_compatibility()
            response = AnalyzeTextResponse(self.post.analyze_text(request_object))
        else:
            raise ValueError(
                "request_object can only be a dictionary or an AnalyzeTextRequest object"
            )
        return response



import pytest

from ..components import *


# File Tests
def test_file_initializer():
    file = File(data="test", content_type="application/pdf")
    assert file.data == "test"
    assert file.content_type == "application/pdf"


def test_file_initializer_fromdict():
    file = File.fromdict({"data": "test", "content_type": "application/pdf"})
    assert file.data == "test"
    assert file.content_type == "application/pdf"


def test_file_invalid_initialize_fromdict():
    error_msg = "File can only accept the values 'data' and 'content_type'"
    with pytest.raises(TypeError) as excinfo:
        File.fromdict(
            {"data": "test", "content_type": "application/pdf", "garbage": "value"}
        )
    assert error_msg in str(excinfo.value)


def test_file_data_validator():
    error_msg = "data must be string-type"
    with pytest.raises(TypeError) as excinfo:
        file = File(data=12, content_type="application/pdf")
    assert error_msg in str(excinfo.value)


def test_file_content_type_validator():
    error_msg = "strawberry/filling is not valid. File.content_type can only be one of the following:"
    with pytest.raises(ValueError) as excinfo:
        file = File(data="test", content_type="strawberry/filling")
    assert error_msg in str(excinfo.value)


def test_file_setters():
    file = File(data="test", content_type="application/pdf")
    file.data = "new test"
    file.content_type = "application/xml"
    assert file.data == "new test"
    assert file.content_type == "application/xml"


# Filter Selector Tests
def test_filter_selector_initializer():
    test_type = "ALLOW"
    test_pattern = "[A-Z]"
    filter_selector = FilterSelector(type=test_type, pattern=test_pattern)
    assert filter_selector.type == test_type
    assert filter_selector.pattern == test_pattern


def test_filter_selector_initializer2():
    test_type = "BLOCK"
    test_pattern = "[A-Z]"
    test_entity_type = "CHARACTER"
    filter_selector = FilterSelector(
        type=test_type, pattern=test_pattern, entity_type=test_entity_type
    )
    assert filter_selector.type == test_type
    assert filter_selector.pattern == test_pattern
    assert filter_selector.entity_type == "CHARACTER"
    assert filter_selector.threshold == 1


def test_filter_selector_initializer():
    test_type = "ALLOW_TEXT"
    test_pattern = "[A-Z]"
    filter_selector = FilterSelector(type=test_type, pattern=test_pattern)
    assert filter_selector.type == test_type
    assert filter_selector.pattern == test_pattern


def test_filter_selector_initialize_fromdict():
    test_type = "ALLOW"
    test_pattern = "[A-Z]"
    test_dict = {"type": test_type, "pattern": test_pattern}
    filter_selector = FilterSelector.fromdict(test_dict)
    assert filter_selector.type == test_type
    assert filter_selector.pattern == test_pattern


def test_filter_selector_invalid_initialize_fromdict():
    with pytest.raises(TypeError) as excinfo:
        FilterSelector.fromdict(
            {"type": "ALLOW", "pattern": "[A-Z]", "fake_key": "fake_value"}
        )
    assert "FilterSelector can only accept the values 'type' and 'pattern'" in str(
        excinfo.value
    )


def test_filter_selector_setters():
    test_type = "ALLOW"
    test_pattern = "[A-Z]"
    filter_selector = FilterSelector(type=test_type, pattern=test_pattern)
    filter_selector.type = "BLOCK"
    filter_selector.pattern = "*1"
    filter_selector.entity_type = "TEST"
    filter_selector.threshold = 0.5

    assert filter_selector.type == "BLOCK"
    assert filter_selector.pattern == "*1"
    assert filter_selector.entity_type == "TEST"
    assert filter_selector.threshold == 0.5


def test_filter_selector_type_validator():
    test_type = "JUNK"
    test_pattern = "[A-Z]"
    with pytest.raises(ValueError) as excinfo:
        FilterSelector(type=test_type, pattern=test_pattern)
    assert (
        " is not valid. FilterSelector.type can only be one of the following: "
        in str(excinfo.value)
    )


def test_filter_selector_pattern_validator():
    test_type = "ALLOW"
    test_pattern = 12
    with pytest.raises(TypeError) as excinfo:
        FilterSelector(type=test_type, pattern=test_pattern)
    assert "FilterSelector.pattern must be of type string" in str(excinfo.value)


def test_filter_selector_entity_type_validator():
    test_type = "BLOCK"
    test_pattern = "[A-Z]"
    test_entity_type = 30
    with pytest.raises(TypeError) as excinfo:
        FilterSelector(
            type=test_type, pattern=test_pattern, entity_type=test_entity_type
        )
    assert "FilterSelector.entity_type must be of type string" in str(excinfo.value)


def test_filter_selector_entity_type_validator():
    test_type = "BLOCK"
    test_pattern = "[A-Z]"
    test_entity_type = "TEST"
    test_threshold = -1
    with pytest.raises(TypeError) as excinfo:
        FilterSelector(
            type=test_type,
            pattern=test_pattern,
            entity_type=test_entity_type,
            threshold=test_threshold,
        )
    assert "FilterSelector.threshold must be greater than 0" in str(excinfo.value)


def test_filter_to_dict():
    test_type = "ALLOW"
    test_pattern = "[A-Z]"
    filter_selector = FilterSelector(type=test_type, pattern=test_pattern)
    filter_dict = filter_selector.to_dict()
    assert filter_dict["type"] == test_type
    assert filter_dict["pattern"] == test_pattern


# Entity Tests
def test_entity_initializer():
    entity = Entity(processed_text="NAME_1", text="this is a test")
    assert entity.processed_text == "NAME_1"
    assert entity.text == "this is a test"


def test_entity_initializer_fromdict():
    entity = Entity.fromdict({"processed_text": "NAME_1", "text": "this is a test"})
    assert entity.processed_text == "NAME_1"
    assert entity.text == "this is a test"


def test_entity_invalid_initialize_fromdict():
    error_msg = "Entity can only accept the values 'processed_text' and 'text'"
    with pytest.raises(TypeError) as excinfo:
        Entity.fromdict(
            {"processed_text": "NAME_1", "text": "this is a test", "garbage": "value"}
        )
    assert error_msg in str(excinfo.value)


def test_entity_processed_text_validator():
    error_msg = "Entity.processed_text must be of type string"
    with pytest.raises(TypeError) as excinfo:
        Entity(processed_text=12, text="ayy")
    assert error_msg in str(excinfo.value)


def test_entity_text_validator():
    error_msg = "Entity.text must be of type string"
    with pytest.raises(TypeError) as excinfo:
        entity = Entity(processed_text="ORGANIZATION_60", text=45.2)
    assert error_msg in str(excinfo.value)


def test_entity_setters():
    entity = Entity(processed_text="NAME_1", text="Jerry Stevens")
    entity.processed_text = "CONDITION_20"
    entity.text = "Broken leg"
    assert entity.processed_text == "CONDITION_20"
    assert entity.text == "Broken leg"


# Entity Type Selector Tests
def test_entity_type_selector_initializer():
    test_type = "DISABLE"

    entity_type_selector = EntityTypeSelector(type=test_type)
    assert entity_type_selector.type == test_type
    assert entity_type_selector.value == []


def test_entity_type_selector_initialize_fromdict():
    entity_type_obj = EntityTypeSelector.fromdict({"type": "ENABLE", "value": ["NAME"]})
    assert entity_type_obj.type == "ENABLE"
    assert entity_type_obj.value == ["NAME"]


def test_entity_type_selector_invalid_initialize_fromdict():
    error_msg = "EntityTypeSelector can only accept the values 'type' and 'value'"
    with pytest.raises(TypeError) as excinfo:
        EntityTypeSelector.fromdict(
            {"type": "ENABLE", "value": ["NAME"], "garbage": "value"}
        )
    assert error_msg in str(excinfo.value)


def test_entity_type_selector_setters():
    entity_type_obj = EntityTypeSelector(type="ENABLE", value=["LOCATION"])
    entity_type_obj.type = "DISABLE"
    assert entity_type_obj.type == "DISABLE"


def test_entity_type_selector_type_validator():
    error_msg = "'JUNK' is not valid. EntityTypeSelector.type can only be one of the following: "
    entity_type_obj = EntityTypeSelector(type="ENABLE", value=["LOCATION"])
    with pytest.raises(ValueError) as excinfo:
        entity_type_obj.type = "JUNK"
    assert error_msg in str(excinfo.value)


def test_entity_type_selector_value_validator():
    error_msg = "EntityTypeSelector.value must be of type list"
    with pytest.raises(TypeError) as excinfo:
        EntityTypeSelector(type="ENABLE", value={})
    assert error_msg in str(excinfo.value)


def test_entity_type_selector_to_dict():
    entity_type_obj = EntityTypeSelector.fromdict(
        {"type": "ENABLE", "value": ["NAME"]}
    ).to_dict()
    assert entity_type_obj["type"] == "ENABLE"
    assert entity_type_obj["value"] == ["NAME"]


# Entity Detection Tests
def test_entity_detection_default_initializer():
    entity_detection = EntityDetection()
    assert entity_detection.accuracy == "high_automatic"
    assert entity_detection.entity_types == []
    assert entity_detection.filter == []
    assert entity_detection.return_entity == True


def test_entity_detection_initializer():
    entity_detection_obj = EntityDetection(
        accuracy="high",
        entity_types=[EntityTypeSelector(type="ENABLE")],
        filter=[FilterSelector(type="ALLOW", pattern="hey")],
        return_entity=False,
    )
    assert entity_detection_obj.accuracy == "high"
    assert type(entity_detection_obj.entity_types[0]) is EntityTypeSelector
    assert type(entity_detection_obj.filter[0]) is FilterSelector
    assert entity_detection_obj.return_entity is False


def test_entity_detection_initialize_fromdict():
    entity_detection = EntityDetection.fromdict(
        {
            "accuracy": "high",
            "entity_types": [EntityTypeSelector(type="ENABLE").to_dict()],
            "filter": [FilterSelector(type="ALLOW", pattern="hey").to_dict()],
            "return_entity": False,
        }
    )
    assert entity_detection.accuracy == "high"
    assert type(entity_detection.entity_types[0]) is EntityTypeSelector
    assert type(entity_detection.filter[0]) is FilterSelector
    assert entity_detection.return_entity is False


def test_entity_detection_invalid_initialize_fromdict():
    error_msg = "EntityDetection can only accept the values 'accuracy', 'entity_types', 'filter' and 'return_entity'"
    with pytest.raises(TypeError) as excinfo:
        EntityDetection.fromdict(
            {
                "accuracy": "high",
                "entity_types": [EntityTypeSelector(type="ENABLE").to_dict()],
                "filter": [FilterSelector(type="ALLOW", pattern="hey").to_dict()],
                "return_entity": False,
                "garbage": "value",
            }
        )
    assert error_msg in str(excinfo.value)


def test_entity_detection_setters():
    entity_detection_obj = EntityDetection.fromdict(
        {
            "accuracy": "high",
            "entity_types": [EntityTypeSelector(type="ENABLE").to_dict()],
            "filter": [FilterSelector(type="ALLOW", pattern="hey").to_dict()],
            "return_entity": False,
        }
    )
    entity_detection_obj.accuracy = "standard"
    entity_detection_obj.return_entity = True
    assert entity_detection_obj.accuracy == "standard"
    assert entity_detection_obj.return_entity == True


def test_entity_detection_accuracy_validator():
    error_msg = (
        "junk is not valid. EntityDetection.accuracy can only be one of the following: "
    )
    entity_detection_obj = EntityDetection.fromdict(
        {
            "accuracy": "high",
            "entity_types": [EntityTypeSelector(type="ENABLE").to_dict()],
            "filter": [FilterSelector(type="ALLOW", pattern="hey").to_dict()],
            "return_entity": False,
        }
    )
    with pytest.raises(ValueError) as excinfo:
        entity_detection_obj.accuracy = "junk"
    assert error_msg in str(excinfo.value)


def test_entity_detection_entity_types_validator():
    error_msg = (
        "EntityDetection.entity_types can only contain EntityTypeSelector objects"
    )
    with pytest.raises(ValueError) as excinfo:
        EntityDetection(
            accuracy="high",
            entity_types=["junk"],
            filter=[FilterSelector(type="ALLOW", pattern="hey")],
            return_entity=False,
        )
    assert error_msg in str(excinfo.value)


def test_entity_detection_entity_types_validator():
    error_msg = "EntityDetection.filter can only contain FilterSelector objects"
    with pytest.raises(ValueError) as excinfo:
        EntityDetection(
            accuracy="high",
            entity_types=[EntityTypeSelector(type="ENABLE")],
            filter=["junk"],
            return_entity=False,
        )
    assert error_msg in str(excinfo.value)


def test_entity_detection_return_entity_validator():
    error_msg = "EntityDetection.return_entity must be of type bool"
    entity_detection_obj = EntityDetection.fromdict(
        {
            "accuracy": "high",
            "entity_types": [EntityTypeSelector(type="ENABLE").to_dict()],
            "filter": [FilterSelector(type="ALLOW", pattern="hey").to_dict()],
            "return_entity": False,
        }
    )
    with pytest.raises(ValueError) as excinfo:
        entity_detection_obj.return_entity = "junk"
    assert error_msg in str(excinfo.value)


def test_entity_detection_to_dict():
    entity_detection_obj = EntityDetection(
        accuracy="high",
        entity_types=[EntityTypeSelector(type="ENABLE")],
        filter=[FilterSelector(type="ALLOW", pattern="hey")],
        return_entity=False,
    ).to_dict()
    assert entity_detection_obj["accuracy"] == "high"
    assert type(entity_detection_obj["entity_types"][0]) is dict
    assert type(entity_detection_obj["filter"][0]) is dict
    assert entity_detection_obj["return_entity"] is False


# Object Entity Type Selector Tests
def test_object_entity_type_selector_initializer():
    test_type = "DISABLE"

    object_entity_type_selector = ObjectEntityTypeSelector(type=test_type)
    assert object_entity_type_selector.type == test_type
    assert object_entity_type_selector.value == []


def test_object_entity_type_selector_initialize_fromdict():
    entity_type_obj = ObjectEntityTypeSelector.fromdict(
        {"type": "ENABLE", "value": ["LOGO"]}
    )
    assert entity_type_obj.type == "ENABLE"
    assert entity_type_obj.value == ["LOGO"]


def test_object_entity_type_selector_invalid_initialize_fromdict():
    error_msg = "ObjectEntityTypeSelector can only accept the values 'type' and 'value'"
    with pytest.raises(TypeError) as excinfo:
        ObjectEntityTypeSelector.fromdict(
            {"type": "ENABLE", "value": ["LOGO"], "garbage": "value"}
        )
    assert error_msg in str(excinfo.value)


def test_object_entity_type_selector_setters():
    entity_type_obj = ObjectEntityTypeSelector(type="ENABLE", value=["FACE"])
    entity_type_obj.type = "DISABLE"
    assert entity_type_obj.type == "DISABLE"


def test_object_entity_type_selector_type_validator():
    error_msg = "'RANDOM' is not valid. ObjectEntityTypeSelector.type can only be one of the following: "
    entity_type_obj = ObjectEntityTypeSelector(type="ENABLE", value=["FACE"])
    with pytest.raises(ValueError) as excinfo:
        entity_type_obj.type = "RANDOM"
    assert error_msg in str(excinfo.value)


def test_object_entity_type_selector_value_validator():
    error_msg = "ObjectEntityTypeSelector.value must be of type list"
    with pytest.raises(TypeError) as excinfo:
        ObjectEntityTypeSelector(type="ENABLE", value={})
    assert error_msg in str(excinfo.value)


def test_object_entity_type_selector_invalid_value_validator():
    error_msg = "ObjectEntityTypeSelector.value can only be one of the following:"
    with pytest.raises(ValueError) as excinfo:
        ObjectEntityTypeSelector(type="ENABLE", value=["RANDOM"])
    assert error_msg in str(excinfo.value)


def test_object_entity_type_selector_to_dict():
    entity_type_obj = ObjectEntityTypeSelector.fromdict(
        {"type": "ENABLE", "value": ["SIGNATURE"]}
    ).to_dict()
    assert entity_type_obj["type"] == "ENABLE"
    assert entity_type_obj["value"] == ["SIGNATURE"]


# Object Entity Detection Tests
def test_object_entity_detection_default_initializer():
    object_entity_detection = ObjectEntityDetection()
    assert object_entity_detection.object_entity_types == []


def test_object_entity_detection_initializer():
    object_entity_detection_obj = ObjectEntityDetection(
        object_entity_types=[ObjectEntityTypeSelector(type="ENABLE")]
    )
    assert (
        type(object_entity_detection_obj.object_entity_types[0])
        is ObjectEntityTypeSelector
    )


def test_object_entity_detection_initialize_fromdict():
    object_entity_detection = ObjectEntityDetection.fromdict(
        {"object_entity_types": [ObjectEntityTypeSelector(type="ENABLE").to_dict()]}
    )
    assert (
        type(object_entity_detection.object_entity_types[0]) is ObjectEntityTypeSelector
    )


def test_object_entity_detection_invalid_initialize_fromdict():
    error_msg = "ObjectEntityDetection can only accept the value 'object_entity_types'"
    with pytest.raises(TypeError) as excinfo:
        ObjectEntityDetection.fromdict(
            {
                "object_entity_types": [
                    ObjectEntityTypeSelector(type="ENABLE").to_dict()
                ],
                "random": "value",
            }
        )
    assert error_msg in str(excinfo.value)


def test_object_entity_detection_object_entity_types_validator():
    error_msg = "ObjectEntityDetection.object_entity_types can only contain ObjectEntityTypeSelector objects"
    with pytest.raises(ValueError) as excinfo:
        ObjectEntityDetection(object_entity_types=["junk"])
    assert error_msg in str(excinfo.value)


def test_object_entity_detection_to_dict():
    object_entity_detection_obj = ObjectEntityDetection(
        object_entity_types=[ObjectEntityTypeSelector(type="ENABLE")]
    ).to_dict()
    assert type(object_entity_detection_obj["object_entity_types"][0]) is dict


# Processed Text Tests
def test_processed_text_default_initializer():
    processed_text = ProcessedText()
    assert processed_text.type == "MARKER"
    assert processed_text.pattern == "[UNIQUE_NUMBERED_ENTITY_TYPE]"
    assert processed_text.coreference_resolution == "heuristics"


def test_processed_text_mask_initializer():
    processed_text = ProcessedText(type="MASK", mask_character="*")
    assert processed_text.type == "MASK"
    assert processed_text.mask_character == "*"


def test_processed_text_coreference_initializer():
    processed_text = ProcessedText(
        type="MARKER", coreference_resolution="model_prediction"
    )
    assert processed_text.type == "MARKER"
    assert processed_text.coreference_resolution == "model_prediction"


def test_processed_text_initialize_fromdict():
    processed_text = ProcessedText.fromdict(
        {
            "type": "MARKER",
            "pattern": "[UNIQUE_NUMBERED_ENTITY_TYPE]",
            "marker_language": "fr",
            "coreference_resolution": "combined",
        }
    )
    assert processed_text.type == "MARKER"
    assert processed_text.pattern == "[UNIQUE_NUMBERED_ENTITY_TYPE]"
    assert processed_text.coreference_resolution == "combined"


def test_processed_text_invalid_initialize_fromdict():
    error_msg = "ProcessedText can only accept the values 'type' and 'pattern'"
    with pytest.raises(TypeError) as excinfo:
        ProcessedText.fromdict(
            {
                "type": "MARKER",
                "pattern": "[UNIQUE_NUMBERED_ENTITY_TYPE]",
                "marker_language": "en",
                "coreference_resolution": "heuristics",
                "junk": "value",
            }
        )
    assert error_msg in str(excinfo.value)


def test_processed_text_mask_setters():
    processed_text = ProcessedText()
    processed_text.type = "MASK"
    processed_text.pattern = "*ALL_ENTITY_TYPES*"
    processed_text.marker_language = "de"
    assert processed_text.type == "MASK"
    assert processed_text.pattern == "*ALL_ENTITY_TYPES*"
    assert processed_text.marker_language == "de"


def test_processed_text_coreference_setters():
    processed_text = ProcessedText()
    processed_text.type = "MARKER"
    processed_text.pattern = "[UNIQUE_NUMBERED_ENTITY_TYPE]"
    processed_text.marker_language = "fr"
    processed_text.coreference_resolution = "model_prediction"
    assert processed_text.type == "MARKER"
    assert processed_text.pattern == "[UNIQUE_NUMBERED_ENTITY_TYPE]"
    assert processed_text.marker_language == "fr"
    assert processed_text.coreference_resolution == "model_prediction"


def test_processed_text_type_validator():
    error_msg = (
        "junk is not valid. ProcessedText.type can only be one of the following: "
    )
    with pytest.raises(ValueError) as excinfo:
        ProcessedText.fromdict(
            {"type": "junk", "pattern": "[UNIQUE_NUMBERED_ENTITY_TYPE]"}
        )
    assert error_msg in str(excinfo.value)


def test_processed_text_pattern_validator():
    error_msg = (
        "junk is not valid. ProcessedText.pattern can only be one of the following: "
    )
    with pytest.raises(ValueError) as excinfo:
        ProcessedText.fromdict({"type": "MARKER", "pattern": "junk"})
    assert error_msg in str(excinfo.value)


def test_processed_text_marker_language_validator():
    error_msg = "junk is not valid. ProcessedText.marker_language can only be one of the following: "
    with pytest.raises(ValueError) as excinfo:
        ProcessedText.fromdict({"type": "MARKER", "marker_language": "junk"})
    assert error_msg in str(excinfo.value)


def test_processed_text_coreference_validator():
    error_msg = "junk is not valid. ProcessedText.coreference_resolution can only be one of the following: "
    with pytest.raises(ValueError) as excinfo:
        ProcessedText.fromdict({"type": "MARKER", "coreference_resolution": "junk"})
    assert error_msg in str(excinfo.value)


def test_processed_text_mask_to_dict():
    processed_text = ProcessedText(type="MASK", mask_character="*").to_dict()
    assert processed_text["type"] == "MASK"
    assert processed_text["mask_character"] == "*"


def test_processed_text_marker_to_dict():
    processed_text = ProcessedText(
        type="MARKER", coreference_resolution="combined"
    ).to_dict()
    assert processed_text["type"] == "MARKER"
    assert processed_text["coreference_resolution"] == "combined"


# PDF Options Tests
def test_pdf_options_default_initializer():
    pdf_options = PDFOptions()
    assert pdf_options.density == 200
    assert pdf_options.max_resolution == 3000
    assert pdf_options.enable_pdf_text_layer is True


def test_pdf_options_initializer():
    pdf_options = PDFOptions(
        density=300, max_resolution=500, enable_pdf_text_layer=False
    )
    assert pdf_options.density == 300
    assert pdf_options.max_resolution == 500
    assert pdf_options.enable_pdf_text_layer is False


def test_pdf_options_initialize_fromdict():
    pdf_options = PDFOptions.fromdict(
        {"density": 300, "max_resolution": 500, "enable_pdf_text_layer": True}
    )
    assert pdf_options.density == 300
    assert pdf_options.max_resolution == 500
    assert pdf_options.enable_pdf_text_layer is True


def test_pdf_options_invalid_initialize_fromdict():
    error_msg = "PDFOptions can only accept 'density', 'max_resolution' and 'enable_pdf_text_layer'"
    with pytest.raises(TypeError) as excinfo:
        PDFOptions.fromdict({"density": 300, "max_resolution": 500, "junk": "value"})
    assert error_msg in str(excinfo.value)


def test_pdf_options_setters():
    pdf_options = PDFOptions(density=300)
    pdf_options.density = 10
    assert pdf_options.density == 10
    pdf_options.max_resolution = 10
    assert pdf_options.max_resolution == 10
    pdf_options.enable_pdf_text_layer = False
    assert pdf_options.enable_pdf_text_layer is False


def test_pdf_options_density_validator():
    error_msg = "PDFOptions.density must be of type int and >0"
    pdf_options = PDFOptions()
    with pytest.raises(ValueError) as excinfo:
        pdf_options.density = "junk"
    assert error_msg in str(excinfo.value)


def test_pdf_options_max_resolution_validator():
    error_msg = "PDFOptions.max_resolution must be of type int and >0"
    pdf_options = PDFOptions()
    with pytest.raises(ValueError) as excinfo:
        pdf_options.max_resolution = "junk"
    assert error_msg in str(excinfo.value)


def test_pdf_options_enable_pdf_text_layer_validator():
    error_msg = "PDFOptions.enable_pdf_text_layer must be of type bool"
    pdf_options = PDFOptions()
    with pytest.raises(ValueError) as excinfo:
        pdf_options.enable_pdf_text_layer = "junk"
    assert error_msg in str(excinfo.value)


def test_pdf_options_to_dict():
    pdf_options = PDFOptions().to_dict()
    assert pdf_options["density"] == 200
    assert pdf_options["max_resolution"] == 3000
    assert pdf_options["enable_pdf_text_layer"] is True


# Audio Options Tests
def test_audio_options_default_initializer():
    audio_options = AudioOptions()
    assert audio_options.bleep_end_padding == 0.5
    assert audio_options.bleep_start_padding == 0.5
    assert audio_options.bleep_frequency == None
    assert audio_options.bleep_gain == None


def test_audio_options_initializer():
    audio_options = AudioOptions(
        bleep_start_padding=200.0,
        bleep_end_padding=300.0,
        bleep_gain=-2,
        bleep_frequency=250,
    )
    assert audio_options.bleep_start_padding == 200.0
    assert audio_options.bleep_end_padding == 300.0
    assert audio_options.bleep_gain == -2
    assert audio_options.bleep_frequency == 250


def test_audio_options_initializer_without_bleep_gain_and_bleep_frequency():
    audio_options = AudioOptions(bleep_start_padding=200.0, bleep_end_padding=300.0)
    assert audio_options.bleep_start_padding == 200.0
    assert audio_options.bleep_end_padding == 300.0
    assert audio_options.bleep_gain == None
    assert audio_options.bleep_frequency == None


def test_audio_options_initialize_fromdict():
    audio_options = AudioOptions.fromdict(
        {
            "bleep_start_padding": 0.3,
            "bleep_end_padding": 0.7,
            "bleep_gain": -2,
            "bleep_frequency": 250,
        }
    )
    assert audio_options.bleep_start_padding == 0.3
    assert audio_options.bleep_end_padding == 0.7
    assert audio_options.bleep_gain == -2
    assert audio_options.bleep_frequency == 250


def test_audio_options_invalid_initialize_fromdict():
    error_msg = "ProcessedText can only accept the values 'bleep_start_padding', 'bleep_end_padding', 'bleep_frequency', and 'bleep_gain'"
    with pytest.raises(TypeError) as excinfo:
        AudioOptions.fromdict(
            {
                "bleep_start_padding": 0.2,
                "bleep_end_padding": 0.1,
                "bleep_frequency": 10,
                "bleep_gain": 0,
                "junk": "value",
            }
        )
    assert error_msg in str(excinfo.value)


def test_audio_options_setters():
    audio_options = AudioOptions()
    audio_options.bleep_end_padding = 1.0
    audio_options.bleep_start_padding = 2.0
    audio_options.bleep_frequency = 100
    audio_options.bleep_gain = -50

    assert audio_options.bleep_end_padding == 1.0
    assert audio_options.bleep_start_padding == 2.0
    assert audio_options.bleep_frequency == 100
    assert audio_options.bleep_gain == -50


def test_audio_options_bleep_start_padding_validator():
    error_msg = "AudioOptions.bleep_start_padding must be of type float"
    with pytest.raises(ValueError) as excinfo:
        AudioOptions().bleep_start_padding = "junk"
    assert error_msg in str(excinfo.value)


def test_audio_options_bleep_end_padding_validator():
    error_msg = "AudioOptions.bleep_end_padding must be of type float"
    with pytest.raises(ValueError) as excinfo:
        AudioOptions().bleep_end_padding = "junk"
    assert error_msg in str(excinfo.value)


def test_audio_options_bleep_gain_validator():
    error_msg = "AudioOptions.bleep_gain must be of type int"
    with pytest.raises(ValueError) as excinfo:
        AudioOptions().bleep_gain = "junk"
    assert error_msg in str(excinfo.value)


def test_audio_options_bleep_frequency_validator():
    error_msg = "AudioOptions.bleep_frequency must be of type int"
    with pytest.raises(ValueError) as excinfo:
        AudioOptions().bleep_frequency = "junk"
    assert error_msg in str(excinfo.value)


def test_audio_options_to_dict():
    audio_options = AudioOptions().to_dict()
    assert audio_options["bleep_end_padding"] == 0.5
    assert audio_options["bleep_start_padding"] == 0.5

    with pytest.raises(KeyError):
        assert audio_options["bleep_gain"] == -3
    with pytest.raises(KeyError):
        assert audio_options["bleep_frequency"] == 600


# Image Options Tests
def test_image_options_default_initializer():
    image_options = ImageOptions()
    assert image_options.masking_method == "blur"
    assert image_options.palette == False


def test_image_options_initializer():
    image_options = ImageOptions(masking_method="blackbox", palette=True)
    assert image_options.masking_method == "blackbox"
    assert image_options.palette == True


def test_image_options_initializer_without_palette():
    image_options = ImageOptions(masking_method="blur")
    assert image_options.masking_method == "blur"
    assert image_options.palette == False


def test_image_options_initialize_fromdict():
    image_options = ImageOptions.fromdict({"masking_method": "blur", "palette": True})
    assert image_options.masking_method == "blur"
    assert image_options.palette == True


def test_image_options_invalid_initialize_fromdict():
    error_msg = "ImageOptions can only accept the values 'masking_method' and 'palette'"
    with pytest.raises(TypeError) as excinfo:
        ImageOptions.fromdict(
            {"masking_method": "blur", "palette": True, "junk": "value"}
        )
    assert error_msg in str(excinfo.value)


def test_image_options_setters():
    image_options = ImageOptions()
    image_options.masking_method = "blur"
    image_options.palette = True

    assert image_options.masking_method == "blur"
    assert image_options.palette == True


def test_image_options_masking_method_validator():
    error_msg = (
        "ImageOptions.masking_method must be one of ['blur', 'blackbox'], but got junk"
    )
    with pytest.raises(ValueError) as excinfo:
        ImageOptions().masking_method = "junk"
    assert error_msg in str(excinfo.value)


def test_image_options_palette_validator():
    error_msg = "ImageOptions.palette must be of type bool, but got <class 'str'>"
    with pytest.raises(ValueError) as excinfo:
        ImageOptions().palette = "junk"
    assert error_msg in str(excinfo.value)


# OCR Options Tests
def test_ocr_options_default_initializer():
    ocr_options = OCROptions()
    assert ocr_options.ocr_system == "paddleocr"


def test_ocr_options_initializer():
    ocr_options = OCROptions(ocr_system="azure_doc_intelligence")
    assert ocr_options.ocr_system == "azure_doc_intelligence"


def test_ocr_options_initialize_fromdict():
    ocr_options = OCROptions.fromdict({"ocr_system": "azure_doc_intelligence"})
    assert ocr_options.ocr_system == "azure_doc_intelligence"


def test_ocr_options_invalid_initialize_fromdict():
    error_msg = "OCROptions can only accept 'ocr_system'"
    with pytest.raises(TypeError) as excinfo:
        OCROptions.fromdict({"ocr_system": "azure_doc_intelligence", "junk": "value"})
    assert error_msg in str(excinfo.value)


def test_ocr_options_setters():
    ocr_options = OCROptions(ocr_system="hybrid")
    assert ocr_options.ocr_system == "hybrid"


def test_ocr_options_ocr_system_validator():
    error_msg = "OCROptions.ocr_system must be one of azure_computer_vision,azure_doc_intelligence,hybrid,paddleocr, but got junk"
    ocr_options = OCROptions()
    with pytest.raises(ValueError) as excinfo:
        ocr_options.ocr_system = "junk"
    assert error_msg in str(excinfo.value)


def test_ocr_options_to_dict():
    ocr_options = OCROptions().to_dict()
    assert ocr_options["ocr_system"] == "paddleocr"


# Timestamp Tests
def test_timestamp_initializer():
    timestamp = Timestamp(start=2.0, end=3.0)
    assert timestamp.start == 2.0
    assert timestamp.end == 3.0


def test_timestamp_initialize_fromdict():
    timestamp = Timestamp.fromdict({"start": 2.0, "end": 3.0})
    assert timestamp.start == 2.0
    assert timestamp.end == 3.0


def test_timestamp_invalid_initialize_fromdict():
    error_msg = "Timestamp can only accept the values 'start' and 'end'"
    with pytest.raises(TypeError) as excinfo:
        Timestamp.fromdict({"start": 2.0, "end": 3.0, "junk": "value"})
    assert error_msg in str(excinfo.value)


def test_timestamp_setters():
    timestamp = Timestamp(start=2.0, end=3.0)
    timestamp.start = 0.0
    timestamp.end = 5.0
    assert timestamp.start == 0.0
    assert timestamp.end == 5.0


def test_timestamp_start_validator():
    error_msg = "Timestamp.start must be of type float"
    with pytest.raises(ValueError) as excinfo:
        Timestamp.fromdict({"start": "test", "end": 3.0})
    assert error_msg in str(excinfo.value)


def test_timestamp_end_validator():
    error_msg = "Timestamp.end must be of type float"
    with pytest.raises(ValueError) as excinfo:
        Timestamp.fromdict({"start": 2.0, "end": "test"})
    assert error_msg in str(excinfo.value)


def test_timestamp_to_dict():
    timestamp = Timestamp(start=2.0, end=3.0).to_dict()
    assert timestamp["start"] == 2.0
    assert timestamp["end"] == 3.0


# Relation Detection Tests
def test_relation_detection_default_initializer():
    relation_detection = RelationDetection()
    assert relation_detection.coreference_resolution is None


def test_relation_detection_initializer():
    relation_detection = RelationDetection(coreference_resolution="model_prediction")
    assert relation_detection.coreference_resolution == "model_prediction"


def test_relation_detection_initialize_fromdict():
    relation_detection = RelationDetection.fromdict(
        {"coreference_resolution": "combined"}
    )
    assert relation_detection.coreference_resolution == "combined"


def test_relation_detection_invalid_initialize_fromdict():
    error_msg = "RelationDetection can only accept the value 'coreference_resolution'"
    with pytest.raises(TypeError) as excinfo:
        RelationDetection.fromdict(
            {"coreference_resolution": "combined", "junk": "value"}
        )
    assert error_msg in str(excinfo.value)


def test_relation_detection_setters():
    relation_detection = RelationDetection()
    relation_detection.coreference_resolution = "combined"
    assert relation_detection.coreference_resolution == "combined"


def test_relation_detection_coreference_resolution_validator():
    error_msg = "junk is not valid. RelationDetection.coreference_resolution can only be one of the following: heuristics, model_prediction, combined or None"
    relation_detection = RelationDetection()
    with pytest.raises(ValueError) as excinfo:
        relation_detection.coreference_resolution = "junk"
    assert error_msg in str(excinfo.value)


def test_relation_detection_to_dict():
    relation_detection = RelationDetection(
        coreference_resolution="model_prediction"
    ).to_dict()
    assert relation_detection["coreference_resolution"] == "model_prediction"


# Process Text Request Tests
def test_process_text_request_default_initializer():
    process_text_request = ProcessTextRequest(text=["hey"])
    assert process_text_request.text == ["hey"]
    assert process_text_request.link_batch is None
    assert process_text_request.entity_detection is None
    assert process_text_request.processed_text is None


def test_process_text_request_initializer():
    text = ["hey!"]
    link_batch = True
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    processed_text = ProcessedText(type="MARKER", pattern="BEST_ENTITY_TYPE")

    process_text_request = ProcessTextRequest(
        text=text,
        link_batch=link_batch,
        entity_detection=entity_detection,
        processed_text=processed_text,
    )

    assert process_text_request.text == text
    assert process_text_request.link_batch == link_batch
    assert process_text_request.entity_detection.accuracy == entity_detection.accuracy
    assert (
        process_text_request.entity_detection.entity_types[0].type == entity_type.type
    )
    assert (
        process_text_request.entity_detection.entity_types[0].value == entity_type.value
    )
    assert process_text_request.entity_detection.filter[0].type == filter.type
    assert process_text_request.entity_detection.filter[0].pattern == filter.pattern
    assert process_text_request.processed_text.type == processed_text.type
    assert process_text_request.processed_text.pattern == processed_text.pattern


def test_process_text_request_initialize_fromdict():
    request_obj = {
        "text": ["hey!"],
        "link_batch": False,
        "entity_detection": {
            "accuracy": "standard",
            "entity_types": [{"type": "DISABLE", "value": ["LOCATION"]}],
            "filter": [{"type": "BLOCK", "pattern": "Roger", "entity_type": "TEST"}],
            "return_entity": False,
        },
        "processed_text": {"type": "MARKER", "pattern": "ALL_ENTITY_TYPES"},
    }
    process_text_request = ProcessTextRequest.fromdict(request_obj)
    assert process_text_request.text == request_obj["text"]
    assert process_text_request.link_batch == request_obj["link_batch"]
    assert (
        process_text_request.entity_detection.accuracy
        == request_obj["entity_detection"]["accuracy"]
    )
    assert (
        process_text_request.entity_detection.entity_types[0].type
        == request_obj["entity_detection"]["entity_types"][0]["type"]
    )
    assert (
        process_text_request.entity_detection.entity_types[0].value
        == request_obj["entity_detection"]["entity_types"][0]["value"]
    )
    assert (
        process_text_request.entity_detection.filter[0].type
        == request_obj["entity_detection"]["filter"][0]["type"]
    )
    assert (
        process_text_request.entity_detection.filter[0].pattern
        == request_obj["entity_detection"]["filter"][0]["pattern"]
    )
    assert (
        process_text_request.processed_text.type
        == request_obj["processed_text"]["type"]
    )
    assert (
        process_text_request.processed_text.pattern
        == request_obj["processed_text"]["pattern"]
    )


def test_process_text_request_invalid_initialize_fromdict():
    error_msg = "ProcessTextRequest can only accept the values 'text', 'link_batch', 'entity_detection' and 'process_text'"
    request_obj = {
        "text": ["hey!"],
        "link_batch": False,
        "entity_detection": {
            "accuracy": "standard",
            "entity_types": [{"type": "DISABLE", "value": ["LOCATION"]}],
            "filter": [{"type": "BLOCK", "pattern": "Roger"}],
            "return_entity": False,
        },
        "processed_text": {"type": "MARKER", "pattern": "ALL_ENTITY_TYPES"},
        "junk": "value",
    }
    with pytest.raises(TypeError) as excinfo:
        ProcessTextRequest.fromdict(request_obj)
    assert error_msg in str(excinfo.value)


def test_process_text_request_to_dict():
    text = ["hey!"]
    link_batch = True
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    processed_text = ProcessedText(type="MARKER", pattern="BEST_ENTITY_TYPE")

    process_text_request = ProcessTextRequest(
        text=text,
        link_batch=link_batch,
        entity_detection=entity_detection,
        processed_text=processed_text,
    ).to_dict()
    assert process_text_request["text"] == text
    assert process_text_request["link_batch"] == link_batch
    assert (
        process_text_request["entity_detection"]["accuracy"]
        == entity_detection.accuracy
    )
    assert (
        process_text_request["entity_detection"]["entity_types"][0]["type"]
        == entity_type.type
    )
    assert (
        process_text_request["entity_detection"]["entity_types"][0]["value"]
        == entity_type.value
    )
    assert process_text_request["entity_detection"]["filter"][0]["type"] == filter.type
    assert (
        process_text_request["entity_detection"]["filter"][0]["pattern"]
        == filter.pattern
    )
    assert process_text_request["processed_text"]["type"] == processed_text.type
    assert process_text_request["processed_text"]["pattern"] == processed_text.pattern


# NER Text Request Tests
def test_ner_text_request_default_initializer():
    ner_text_request = NerTextRequest(text=["hey"])
    assert ner_text_request.text == ["hey"]
    assert ner_text_request.link_batch is None
    assert ner_text_request.entity_detection is None


def test_ner_text_request_initializer():
    text = ["hey!"]
    link_batch = True
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )

    ner_text_request = NerTextRequest(
        text=text, link_batch=link_batch, entity_detection=entity_detection
    )

    assert ner_text_request.text == text
    assert ner_text_request.link_batch == link_batch
    assert ner_text_request.entity_detection.accuracy == entity_detection.accuracy
    assert ner_text_request.entity_detection.entity_types[0].type == entity_type.type
    assert ner_text_request.entity_detection.entity_types[0].value == entity_type.value
    assert ner_text_request.entity_detection.filter[0].type == filter.type
    assert ner_text_request.entity_detection.filter[0].pattern == filter.pattern


def test_ner_text_request_initialize_fromdict():
    request_obj = {
        "text": ["hey!"],
        "link_batch": False,
        "entity_detection": {
            "accuracy": "standard",
            "entity_types": [{"type": "DISABLE", "value": ["LOCATION"]}],
            "filter": [{"type": "BLOCK", "pattern": "Roger", "entity_type": "TEST"}],
            "return_entity": False,
        },
    }
    ner_text_request = NerTextRequest.fromdict(request_obj)
    assert ner_text_request.text == request_obj["text"]
    assert ner_text_request.link_batch == request_obj["link_batch"]
    assert (
        ner_text_request.entity_detection.accuracy
        == request_obj["entity_detection"]["accuracy"]
    )
    assert (
        ner_text_request.entity_detection.entity_types[0].type
        == request_obj["entity_detection"]["entity_types"][0]["type"]
    )
    assert (
        ner_text_request.entity_detection.entity_types[0].value
        == request_obj["entity_detection"]["entity_types"][0]["value"]
    )
    assert (
        ner_text_request.entity_detection.filter[0].type
        == request_obj["entity_detection"]["filter"][0]["type"]
    )
    assert (
        ner_text_request.entity_detection.filter[0].pattern
        == request_obj["entity_detection"]["filter"][0]["pattern"]
    )


def test_ner_text_request_invalid_initialize_fromdict():
    error_msg = "NerTextRequest can only accept the values 'text', 'link_batch' and 'entity_detection'"
    request_obj = {
        "text": ["hey!"],
        "link_batch": False,
        "entity_detection": {
            "accuracy": "standard",
            "entity_types": [{"type": "DISABLE", "value": ["LOCATION"]}],
            "filter": [{"type": "BLOCK", "pattern": "Roger"}],
            "return_entity": False,
        },
        "junk": "value",
    }
    with pytest.raises(TypeError) as excinfo:
        NerTextRequest.fromdict(request_obj)
    assert error_msg in str(excinfo.value)


def test_ner_text_request_to_dict():
    text = ["hey!"]
    link_batch = True
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )

    ner_text_request = NerTextRequest(
        text=text, link_batch=link_batch, entity_detection=entity_detection
    ).to_dict()
    assert ner_text_request["text"] == text
    assert ner_text_request["link_batch"] == link_batch
    assert ner_text_request["entity_detection"]["accuracy"] == entity_detection.accuracy
    assert (
        ner_text_request["entity_detection"]["entity_types"][0]["type"]
        == entity_type.type
    )
    assert (
        ner_text_request["entity_detection"]["entity_types"][0]["value"]
        == entity_type.value
    )
    assert ner_text_request["entity_detection"]["filter"][0]["type"] == filter.type
    assert (
        ner_text_request["entity_detection"]["filter"][0]["pattern"] == filter.pattern
    )


# Analyze Text Request Tests
def test_analyze_text_request_default_initializer():
    analyze_text_request = AnalyzeTextRequest(text=["hey"], locale="en-US")
    assert analyze_text_request.text == ["hey"]
    assert analyze_text_request.locale == "en-US"
    assert analyze_text_request.link_batch is None
    assert analyze_text_request.entity_detection is None
    assert analyze_text_request.relation_detection is None


def test_analyze_text_request_initializer():
    text = ["hey!"]
    locale = "en-US"
    link_batch = True
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    relation_detection = RelationDetection(coreference_resolution="model_prediction")

    analyze_text_request = AnalyzeTextRequest(
        text=text,
        locale=locale,
        link_batch=link_batch,
        entity_detection=entity_detection,
        relation_detection=relation_detection,
    )

    assert analyze_text_request.text == text
    assert analyze_text_request.locale == locale
    assert analyze_text_request.link_batch == link_batch
    assert analyze_text_request.entity_detection.accuracy == entity_detection.accuracy
    assert (
        analyze_text_request.entity_detection.entity_types[0].type == entity_type.type
    )
    assert (
        analyze_text_request.entity_detection.entity_types[0].value == entity_type.value
    )
    assert analyze_text_request.entity_detection.filter[0].type == filter.type
    assert analyze_text_request.entity_detection.filter[0].pattern == filter.pattern
    assert (
        analyze_text_request.relation_detection.coreference_resolution
        == relation_detection.coreference_resolution
    )


def test_analyze_text_request_initialize_fromdict():
    request_obj = {
        "text": ["hey!"],
        "locale": "en-US",
        "link_batch": False,
        "entity_detection": {
            "accuracy": "standard",
            "entity_types": [{"type": "DISABLE", "value": ["LOCATION"]}],
            "filter": [{"type": "BLOCK", "pattern": "Roger", "entity_type": "TEST"}],
            "return_entity": False,
        },
        "relation_detection": {"coreference_resolution": "combined"},
    }
    analyze_text_request = AnalyzeTextRequest.fromdict(request_obj)
    assert analyze_text_request.text == request_obj["text"]
    assert analyze_text_request.locale == request_obj["locale"]
    assert analyze_text_request.link_batch == request_obj["link_batch"]
    assert (
        analyze_text_request.entity_detection.accuracy
        == request_obj["entity_detection"]["accuracy"]
    )
    assert (
        analyze_text_request.entity_detection.entity_types[0].type
        == request_obj["entity_detection"]["entity_types"][0]["type"]
    )
    assert (
        analyze_text_request.entity_detection.entity_types[0].value
        == request_obj["entity_detection"]["entity_types"][0]["value"]
    )
    assert (
        analyze_text_request.entity_detection.filter[0].type
        == request_obj["entity_detection"]["filter"][0]["type"]
    )
    assert (
        analyze_text_request.entity_detection.filter[0].pattern
        == request_obj["entity_detection"]["filter"][0]["pattern"]
    )
    assert (
        analyze_text_request.relation_detection.coreference_resolution
        == request_obj["relation_detection"]["coreference_resolution"]
    )


def test_analyze_text_request_invalid_initialize_fromdict():
    error_msg = "AnalyzeTextRequest can only accept the values 'text', 'locale', 'link_batch', 'entity_detection' and 'relation_detection'"
    request_obj = {
        "text": ["hey!"],
        "link_batch": False,
        "entity_detection": {
            "accuracy": "standard",
            "entity_types": [{"type": "DISABLE", "value": ["LOCATION"]}],
            "filter": [{"type": "BLOCK", "pattern": "Roger"}],
            "return_entity": False,
        },
        "relation_detection": {"coreference_resolution": "model_prediction"},
        "junk": "value",
    }
    with pytest.raises(TypeError) as excinfo:
        AnalyzeTextRequest.fromdict(request_obj)
    assert error_msg in str(excinfo.value)


def test_analyze_text_request_to_dict():
    text = ["hey!"]
    locale = "en"
    link_batch = True
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    relation_detection = RelationDetection(coreference_resolution="heuristics")

    analyze_text_request = AnalyzeTextRequest(
        text=text,
        locale=locale,
        link_batch=link_batch,
        entity_detection=entity_detection,
        relation_detection=relation_detection,
    ).to_dict()
    assert analyze_text_request["text"] == text
    assert analyze_text_request["locale"] == locale
    assert analyze_text_request["link_batch"] == link_batch
    assert (
        analyze_text_request["entity_detection"]["accuracy"]
        == entity_detection.accuracy
    )
    assert (
        analyze_text_request["entity_detection"]["entity_types"][0]["type"]
        == entity_type.type
    )
    assert (
        analyze_text_request["entity_detection"]["entity_types"][0]["value"]
        == entity_type.value
    )
    assert analyze_text_request["entity_detection"]["filter"][0]["type"] == filter.type
    assert (
        analyze_text_request["entity_detection"]["filter"][0]["pattern"]
        == filter.pattern
    )
    assert (
        analyze_text_request["relation_detection"]["coreference_resolution"]
        == "heuristics"
    )


# Process File URI Request Tests
def test_process_file_uri_request_default_initializer():
    process_file_uri_obj = ProcessFileUriRequest(uri="this/location/right/here.png")
    assert process_file_uri_obj.uri == "this/location/right/here.png"
    assert process_file_uri_obj.entity_detection is None
    assert process_file_uri_obj.object_entity_detection is None
    assert process_file_uri_obj.pdf_options is None
    assert process_file_uri_obj.audio_options is None


def test_process_file_uri_request_initializer():
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    pdf_options = PDFOptions(density=100)
    audio_options = AudioOptions(
        bleep_start_padding=1.0,
        bleep_end_padding=2.0,
        bleep_frequency=200,
        bleep_gain=-2,
    )
    object_entity_type = ObjectEntityTypeSelector(type="ENABLE", value=["LOGO"])
    object_entity_detection = ObjectEntityDetection(
        object_entity_types=[object_entity_type]
    )
    process_file_uri_obj = ProcessFileUriRequest(
        uri="this/location/right/here.png",
        entity_detection=entity_detection,
        object_entity_detection=object_entity_detection,
        pdf_options=pdf_options,
        audio_options=audio_options,
    )
    assert process_file_uri_obj.uri == "this/location/right/here.png"
    assert process_file_uri_obj.entity_detection.accuracy == "standard"
    assert (
        process_file_uri_obj.object_entity_detection.object_entity_types[0].type
        == "ENABLE"
    )
    assert process_file_uri_obj.pdf_options.density == 100
    assert process_file_uri_obj.audio_options.bleep_end_padding == 2.0
    assert process_file_uri_obj.audio_options.bleep_frequency == 200
    assert process_file_uri_obj.audio_options.bleep_gain == -2


def test_process_file_uri_request_initialize_fromdict():
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    pdf_options = PDFOptions(density=100)
    audio_options = AudioOptions(
        bleep_start_padding=1.0,
        bleep_end_padding=2.0,
        bleep_frequency=200,
        bleep_gain=-2,
    )
    object_entity_type = ObjectEntityTypeSelector(type="ENABLE", value=["LOGO"])
    object_entity_detection = ObjectEntityDetection(
        object_entity_types=[object_entity_type]
    )
    ocr_options = OCROptions(ocr_system="azure_computer_vision")
    process_file_uri_obj = ProcessFileUriRequest.fromdict(
        {
            "uri": "this/location/right/here.png",
            "entity_detection": entity_detection.to_dict(),
            "object_entity_detection": object_entity_detection.to_dict(),
            "pdf_options": pdf_options.to_dict(),
            "audio_options": audio_options.to_dict(),
            "ocr_options": ocr_options.to_dict(),
        }
    )
    assert process_file_uri_obj.uri == "this/location/right/here.png"
    assert process_file_uri_obj.entity_detection.accuracy == "standard"
    assert (
        process_file_uri_obj.object_entity_detection.object_entity_types[0].type
        == "ENABLE"
    )
    assert process_file_uri_obj.pdf_options.density == 100
    assert process_file_uri_obj.audio_options.bleep_end_padding == 2.0
    assert process_file_uri_obj.audio_options.bleep_frequency == 200
    assert process_file_uri_obj.audio_options.bleep_gain == -2
    assert process_file_uri_obj.ocr_options.ocr_system == "azure_computer_vision"


def test_process_file_uri_request_invalid_initialize_fromdict():
    error_msg = "ProcessFileUriRequest can only accept the values 'uri', 'entity_detection', 'object_entity_detection', 'pdf_options', 'audio_options', 'image_options' and 'ocr_options'"
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    pdf_options = PDFOptions(density=100)
    audio_options = AudioOptions(
        bleep_start_padding=1.0,
        bleep_end_padding=2.0,
        bleep_frequency=200,
        bleep_gain=-2,
    )
    object_entity_type = ObjectEntityTypeSelector(type="ENABLE", value=["LOGO"])
    object_entity_detection = ObjectEntityDetection(
        object_entity_types=[object_entity_type]
    )
    ocr_options = OCROptions(ocr_system="azure_computer_vision")
    with pytest.raises(TypeError) as excinfo:
        ProcessFileUriRequest.fromdict(
            {
                "uri": "this/location/right/here.png",
                "entity_detection": entity_detection.to_dict(),
                "object_entity_detection": object_entity_detection.to_dict(),
                "pdf_options": pdf_options.to_dict(),
                "audio_options": audio_options.to_dict(),
                "ocr_options": ocr_options.to_dict(),
                "junk": "value",
            }
        )
    assert error_msg in str(excinfo.value)


def test_process_file_uri_request_to_dict():
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    pdf_options = PDFOptions(density=100)
    audio_options = AudioOptions(
        bleep_start_padding=1.0,
        bleep_end_padding=2.0,
        bleep_frequency=200,
        bleep_gain=-2,
    )
    object_entity_type = ObjectEntityTypeSelector(type="ENABLE", value=["LOGO"])
    object_entity_detection = ObjectEntityDetection(
        object_entity_types=[object_entity_type]
    )
    process_file_uri_obj = ProcessFileUriRequest(
        uri="this/location/right/here.png",
        entity_detection=entity_detection,
        object_entity_detection=object_entity_detection,
        pdf_options=pdf_options,
        audio_options=audio_options,
    ).to_dict()
    assert process_file_uri_obj["uri"] == "this/location/right/here.png"
    assert process_file_uri_obj["entity_detection"]["accuracy"] == "standard"
    assert (
        process_file_uri_obj["object_entity_detection"]["object_entity_types"][0][
            "type"
        ]
        == "ENABLE"
    )
    assert process_file_uri_obj["pdf_options"]["density"] == 100
    assert process_file_uri_obj["audio_options"]["bleep_end_padding"] == 2.0
    assert process_file_uri_obj["audio_options"]["bleep_frequency"] == 200
    assert process_file_uri_obj["audio_options"]["bleep_gain"] == -2


# Process File Base64 Request Tests
def test_process_file_base64_request_default_initializer():
    process_file_base64_request_obj = ProcessFileBase64Request(
        file="sfsfxe234jkjsdlkfnDATA!!!!!!"
    )
    assert process_file_base64_request_obj.file == "sfsfxe234jkjsdlkfnDATA!!!!!!"
    assert process_file_base64_request_obj.entity_detection is None
    assert process_file_base64_request_obj.object_entity_detection is None
    assert process_file_base64_request_obj.pdf_options is None
    assert process_file_base64_request_obj.audio_options is None


def test_process_file_base64_request_initializer():
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    pdf_options = PDFOptions(density=100)
    audio_options = AudioOptions(
        bleep_start_padding=1.0,
        bleep_end_padding=2.0,
        bleep_frequency=200,
        bleep_gain=-2,
    )
    object_entity_type = ObjectEntityTypeSelector(type="ENABLE", value=["LOGO"])
    object_entity_detection = ObjectEntityDetection(
        object_entity_types=[object_entity_type]
    )
    process_file_base64_request_obj = ProcessFileBase64Request(
        file="sfsfxe234jkjsdlkfnDATA",
        entity_detection=entity_detection,
        object_entity_detection=object_entity_detection,
        pdf_options=pdf_options,
        audio_options=audio_options,
    )
    assert process_file_base64_request_obj.file == "sfsfxe234jkjsdlkfnDATA"
    assert process_file_base64_request_obj.entity_detection.accuracy == "standard"
    assert (
        process_file_base64_request_obj.object_entity_detection.object_entity_types[
            0
        ].type
        == "ENABLE"
    )
    assert process_file_base64_request_obj.pdf_options.density == 100
    assert process_file_base64_request_obj.audio_options.bleep_end_padding == 2.0
    assert process_file_base64_request_obj.audio_options.bleep_frequency == 200
    assert process_file_base64_request_obj.audio_options.bleep_gain == -2


def test_process_file_base64_request_initialize_fromdict():
    file = File(data="sfsfxe234jkjsdlkfnDATA", content_type="application/pdf")
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    pdf_options = PDFOptions(density=100)
    audio_options = AudioOptions(
        bleep_start_padding=1.0,
        bleep_end_padding=2.0,
        bleep_frequency=200,
        bleep_gain=-2,
    )
    object_entity_type = ObjectEntityTypeSelector(type="ENABLE", value=["LOGO"])
    object_entity_detection = ObjectEntityDetection(
        object_entity_types=[object_entity_type]
    )
    process_file_base64_request_obj = ProcessFileBase64Request.fromdict(
        {
            "file": file.to_dict(),
            "entity_detection": entity_detection.to_dict(),
            "object_entity_detection": object_entity_detection.to_dict(),
            "pdf_options": pdf_options.to_dict(),
            "audio_options": audio_options.to_dict(),
        }
    )
    assert process_file_base64_request_obj.file.data == "sfsfxe234jkjsdlkfnDATA"
    assert process_file_base64_request_obj.entity_detection.accuracy == "standard"
    assert (
        process_file_base64_request_obj.object_entity_detection.object_entity_types[
            0
        ].type
        == "ENABLE"
    )
    assert process_file_base64_request_obj.pdf_options.density == 100
    assert process_file_base64_request_obj.audio_options.bleep_end_padding == 2.0
    assert process_file_base64_request_obj.audio_options.bleep_frequency == 200
    assert process_file_base64_request_obj.audio_options.bleep_gain == -2


def test_process_file_base64_request_invalid_initialize_fromdict():
    error_msg = "ProcessFileBase64Request can only accept the values 'file', 'entity_detection', 'object_entity_detection', 'pdf_options', 'audio_options', 'image_options' and 'ocr_options'"
    file = File(data="sfsfxe234jkjsdlkfnDATA", content_type="application/pdf")
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    pdf_options = PDFOptions(density=100)
    audio_options = AudioOptions(
        bleep_start_padding=1.0,
        bleep_end_padding=2.0,
        bleep_frequency=200,
        bleep_gain=-2,
    )
    object_entity_type = ObjectEntityTypeSelector(type="ENABLE", value=["LOGO"])
    object_entity_detection = ObjectEntityDetection(
        object_entity_types=[object_entity_type]
    )
    ocr_options = OCROptions(ocr_system="azure_computer_vision")
    with pytest.raises(TypeError) as excinfo:
        ProcessFileBase64Request.fromdict(
            {
                "file": file.to_dict(),
                "entity_detection": entity_detection.to_dict(),
                "object_entity_detection": object_entity_detection.to_dict(),
                "pdf_options": pdf_options.to_dict(),
                "audio_options": audio_options.to_dict(),
                "ocr_options": ocr_options.to_dict(),
                "junk": "value",
            }
        )
    assert error_msg in str(excinfo.value)


def test_process_file_base64_request_to_dict():
    entity_type = EntityTypeSelector(type="ENABLE", value=["NAME"])
    filter = FilterSelector(type="ALLOW", pattern="hey")
    entity_detection = EntityDetection(
        accuracy="standard",
        entity_types=[entity_type],
        filter=[filter],
        return_entity=False,
    )
    pdf_options = PDFOptions(density=100)
    audio_options = AudioOptions(
        bleep_start_padding=1.0,
        bleep_end_padding=2.0,
        bleep_frequency=200,
        bleep_gain=-2,
    )
    object_entity_type = ObjectEntityTypeSelector(type="DISABLE", value=["LOGO"])
    object_entity_detection = ObjectEntityDetection(
        object_entity_types=[object_entity_type]
    )
    process_file_base64_request_obj = ProcessFileBase64Request(
        file="sfsfxe234jkjsdlkfnDATA",
        entity_detection=entity_detection,
        object_entity_detection=object_entity_detection,
        pdf_options=pdf_options,
        audio_options=audio_options,
    ).to_dict()
    assert process_file_base64_request_obj["file"] == "sfsfxe234jkjsdlkfnDATA"
    assert process_file_base64_request_obj["entity_detection"]["accuracy"] == "standard"
    assert (
        process_file_base64_request_obj["object_entity_detection"][
            "object_entity_types"
        ][0]["type"]
        == "DISABLE"
    )
    assert process_file_base64_request_obj["pdf_options"]["density"] == 100
    assert process_file_base64_request_obj["audio_options"]["bleep_end_padding"] == 2.0
    assert process_file_base64_request_obj["audio_options"]["bleep_frequency"] == 200
    assert process_file_base64_request_obj["audio_options"]["bleep_gain"] == -2


# Bleep Request Tests
def test_bleep_request_initializer():
    file = File(data="test", content_type="image/jpg")
    timestamps = [Timestamp(start=0.0, end=1.0)]
    bleep_request = BleepRequest(file=file, timestamps=timestamps)
    assert bleep_request.file.data == "test"
    assert bleep_request.timestamps[0].start == 0


def test_bleep_request_initialize_fromdict():
    file = File(data="test", content_type="image/jpg")
    timestamps = [Timestamp(start=0.0, end=1.0)]
    bleep_request = BleepRequest.fromdict(
        {"file": file.to_dict(), "timestamps": [row.to_dict() for row in timestamps]}
    )
    assert bleep_request.file.data == "test"
    assert bleep_request.timestamps[0].start == 0


def test_bleep_request_invalid_initialize_fromdict():
    error_msg = "BleepRequest can only accept the values 'file', 'timestamps', 'bleep_frequency', and 'bleep_gain'"
    file = File(data="test", content_type="image/jpg")
    timestamps = [Timestamp(start=0.0, end=1.0)]
    with pytest.raises(TypeError) as excinfo:
        BleepRequest.fromdict(
            {
                "file": file.to_dict(),
                "timestamps": [row.to_dict() for row in timestamps],
                "bleep_gain": 2,
                "bleep_frequency": 400,
                "junk": "value",
            }
        )
    assert error_msg in str(excinfo.value)


def test_bleep_request_to_dict():
    file = File(data="test", content_type="image/jpg")
    timestamps = [Timestamp(start=0.0, end=1.0)]
    bleep_request = BleepRequest(file=file, timestamps=timestamps).to_dict()
    assert bleep_request["file"]["data"] == "test"
    assert bleep_request["timestamps"][0]["start"] == 0


# Reidentify Text Request Tests
def test_reidentify_text_request_initializer():
    processed_text = ["this is a test"]
    entities = [Entity(processed_text="NAME", text="Hola")]
    model = "gpt-700.2-ultra-turbo"
    reid = ReidentifyTextRequest(
        processed_text=processed_text,
        entities=entities,
        model=model,
        reidentify_sensitive_fields=True,
    )
    assert reid.processed_text == ["this is a test"]
    assert reid.entities[0].text == "Hola"
    assert reid.model == "gpt-700.2-ultra-turbo"
    assert reid.reidentify_sensitive_fields == True


def test_reidentify_text_request_initialize_fromdict():
    processed_text = ["this is a test"]
    entities = [Entity(processed_text="NAME", text="Hola")]
    model = "gpt-700.2-ultra-turbo"
    reid = ReidentifyTextRequest.fromdict(
        {
            "processed_text": processed_text,
            "entities": [row.to_dict() for row in entities],
            "model": model,
            "reidentify_sensitive_fields": False,
        }
    )
    assert reid.processed_text == ["this is a test"]
    assert reid.entities[0].text == "Hola"
    assert reid.model == "gpt-700.2-ultra-turbo"
    assert reid.reidentify_sensitive_fields == False


def test_reidentify_text_request_invalid_initialize_fromdict():
    error_msg = "ReidentifyTextRequest can only accept the values 'processed_text', 'entities', 'model' and 'reidentify_sensitive_fields'"
    processed_text = ["this is a test"]
    entities = [Entity(processed_text="NAME", text="Hola")]
    model = "gpt-700.2-ultra-turbo"
    with pytest.raises(TypeError) as excinfo:
        ReidentifyTextRequest.fromdict(
            {
                "processed_text": processed_text,
                "entities": [row.to_dict() for row in entities],
                "model": model,
                "junk": "value",
            }
        )
        assert error_msg in str(excinfo.value)


def test_reidentify_text_request_to_dict():
    processed_text = ["this is a test"]
    entities = [Entity(processed_text="NAME", text="Hola")]
    model = "gpt-700.2-ultra-turbo"
    reidentify_sensitive_fields = False
    reid = ReidentifyTextRequest(
        processed_text=processed_text,
        entities=entities,
        model=model,
        reidentify_sensitive_fields=reidentify_sensitive_fields,
    ).to_dict()
    assert reid["processed_text"][0] == "this is a test"
    assert reid["entities"][0]["text"] == "Hola"
    assert reid["model"] == "gpt-700.2-ultra-turbo"
    assert reid["reidentify_sensitive_fields"] == reidentify_sensitive_fields


def test_synthetic_text():
    processed_text = ProcessedText(
        type="SYNTHETIC",
        synthetic_entity_accuracy="standard",
        preserve_relationships=True,
    )
    assert processed_text.type == "SYNTHETIC"
    assert hasattr(processed_text, "pattern") == False


def test_change_type():
    processed_text = ProcessedText(
        type="MARKER", pattern="UNIQUE_NUMBERED_ENTITY_TYPE"
    )  # Marker
    processed_text.type = "MASK"
    assert processed_text.type == "MASK"
    assert hasattr(processed_text, "pattern") == False
    assert processed_text.mask_character == "#"


def test_mask_invalid_mask_character_initialize():
    error_msg = "mask_character must have only one character."
    with pytest.raises(ValueError) as excinfo:
        processed_text = ProcessedText(type="MASK", mask_character="invalid")
    assert error_msg in str(excinfo.value)


def test_mask_invalid_mask_character():
    error_msg = "mask_character must have only one character."
    processed_text = ProcessedText()
    processed_text.type = "MASK"
    assert processed_text.type == "MASK"
    with pytest.raises(ValueError) as excinfo:
        processed_text.mask_character = "invalid"
    assert error_msg in str(excinfo.value)


def test_synthetic_invalid_accuracy_initialize():
    error_msg = "Synthetic Entity Accuracy can only accept values"
    with pytest.raises(ValueError) as excinfo:
        processed_text = ProcessedText(
            type="SYNTHETIC",
            synthetic_entity_accuracy="invalid",
            preserve_relationships=True,
        )
    assert error_msg in str(excinfo.value)


def test_synthetic_invalid_accuracy():
    error_msg = "Synthetic Entity Accuracy can only accept values"
    processed_text = ProcessedText(type="SYNTHETIC", preserve_relationships=True)
    with pytest.raises(ValueError) as excinfo:
        processed_text.synthetic_entity_accuracy = "invalid"
    assert error_msg in str(excinfo.value)



from typing import Union

import requests

from .pai_uris import PAIURIs


class PAIRequests:
    def __init__(self, uris: PAIURIs):
        self._uris = uris
        self.headers = self.base_header

    @property
    def uris(self):
        return self._uris

    @property
    def base_header(self):
        return {"Accept": "application/json"}

    def make_request(
        self,
        request_type: Union[requests.get, requests.post],
        uri: str,
        payload: dict = None,
    ):
        response = request_type(uri, json=payload, headers=self.headers)
        return response


class PAIGetRequests(PAIRequests):
    def __init__(self, uris: PAIURIs):
        """
        A class of get requests used by the client
        """
        self.request_type = requests.get
        super(PAIGetRequests, self).__init__(uris)

    def health(self):
        return self.make_request(self.request_type, self.uris.health)

    def metrics(self):
        return self.make_request(self.request_type, self.uris.metrics)

    def version(self):
        return self.make_request(self.request_type, self.uris.version)

    def diagnostics(self):
        return self.make_request(self.request_type, self.uris.diagnostics)


class PAIPostRequests(PAIRequests):
    def __init__(self, uris: PAIURIs):
        self.request_type = requests.post
        super(PAIPostRequests, self).__init__(uris)

    def process_text(self, request_object):
        return self.make_request(
            self.request_type, self.uris.process_text, request_object
        )

    def process_files_uri(self, request_object):
        return self.make_request(
            self.request_type, self.uris.process_files_uri, request_object
        )

    def process_files_base64(self, request_object):
        return self.make_request(
            self.request_type, self.uris.process_files_base64, request_object
        )

    def bleep(self, request_object):
        return self.make_request(self.request_type, self.uris.bleep, request_object)

    def reidentify_text(self, request_object):
        return self.make_request(
            self.request_type, self.uris.reidentify_text, request_object
        )

    def ner_text(self, request_object):
        return self.make_request(self.request_type, self.uris.ner_text, request_object)

    def analyze_text(self, request_object):
        return self.make_request(
            self.request_type, self.uris.analyze_text, request_object
        )
